"""
Map positions on physical focal plane to sky using physical optics.
Also includes tools for computing transforms between point clouds.

LAT code adapted from code provided by Simon Dicker.
"""
import logging
from functools import lru_cache, partial
import numpy as np
from scipy.interpolate import make_interp_spline, bisplrep, bisplev
from scipy.spatial.transform import Rotation as R
from sotodlib import core
from sotodlib.io.metadata import read_dataset
from so3g.proj import quat
import yaml

logger = logging.getLogger(__name__)

# Dictionary of zemax tube layout.
# +ve x is to the right as seen from back of the cryostat *need to check*
#           11
#    5    3    4    6
#       1    0    2
#    9    7    8    10
#           12
# Below config assumes a 30 degree rotation
LAT_TUBES = {
    "c1": 0,
    "i1": 3,
    "i2": 1,
    "i3": 7,
    "i4": 8,
    "i5": 2,
    "i6": 4,
    "o1": 5,
    "o2": 9,
    "o3": 12,
    "o4": 10,
    "o5": 6,
    "o6": 11,
}

# fmt: off
# SAT Optics
# TODO: Maybe we want these to be provided in a config file?
SAT_R_FP = (0.0, 29.7580, 59.4574, 89.5745, 120.550, 152.821, 163.986, 181.218)
SAT_R_SKY = (0.0, 0.0523597, 0.10471958, 0.15707946, 0.20943951, 0.26179764, 0.27925093, 0.30543087)
# fmt: on

def _interp_func(x, y, spline):
    xr = np.atleast_1d(x).ravel()
    xa = np.argsort(xr)
    xs = np.argsort(xa)
    yr = np.atleast_1d(y).ravel()
    ya = np.argsort(yr)
    ys = np.argsort(ya)
    z = bisplev(xr[xa], yr[ya], spline)

    if np.isscalar(z):
        if np.isscalar(x):
            return z
        else:
            z = np.atleast_2d(z)
    z = z[(xs, ys)]
    if np.isscalar(x):
        return z[0]

    return z.reshape(x.shape)


def gen_pol_endpoints(x, y, pol):
    """
    Get end points of unit vectors that will be centered on the provided xy positions
    and have the angles specified by pol.

    Arguments:

        x: X positions nominally in mm.

        y: Y positions nominally in mm.

        pol: Angles in degrees.

    Returns:

        x_pol: X postions of endpoints where the even indices are starts
        and the odds are ends.

        y_pol: Y postions of endpoints where the even indices are starts
        and the odds are ends.
    """
    _x = np.cos(np.deg2rad(pol)) / 2
    _y = np.sin(np.deg2rad(pol)) / 2

    x_pol = np.column_stack((x - _x, x + _x)).ravel()
    y_pol = np.column_stack((y - _y, y + _y)).ravel()

    return x_pol, y_pol


def get_gamma(pol_xi, pol_eta):
    """
    Convert xi, eta endpoints to angles that correspond to gamma.

    Arguments:

        pol_xi: 2n xi values where the ones with even indices are
                starting points and odd are ending points where each pair
                forms a vector whose angle is gamma.

        pol_eta: 2n eta values where the ones with even indices are
                 starting points and odd are ending points where each pair
                 forms a vector whose angle is gamma.

    Returns:

        gamma: Array of n gamma values in radians.
    """
    xi = pol_xi.reshape((-1, 2))
    eta = pol_eta.reshape((-1, 2))

    q0 = quat.rotation_xieta(xi[:,0], eta[:,0])
    q1 = quat.rotation_xieta(xi[:,1], eta[:,1])
    dq = ~q0 * q1
    d_xi, d_eta, _ = quat.decompose_xieta(dq)

    gamma = np.arctan2(d_xi, d_eta) % (2 * np.pi)
    return gamma


@lru_cache(maxsize=None)
def load_ufm_to_fp_config(config_path):
    """
    Load and cache config file with the parameters to transform from UFM to focal_plane coordinates.

    Arguments:

        config_path: Path to the yaml config file.

    Returns:

        config: Dictionairy containing the config information.
    """
    with open(config_path, "r") as file:
        config = yaml.safe_load(file)
    return config


@lru_cache(maxsize=None)
def get_ufm_to_fp_pars(telescope_flavor, wafer_slot, config_path):
    """
    Get (and cache) the parameters to transform from UFM to focal plane coordinates
    for a specific slot of a given telescope's focal plane.

    Arguments:

        telescope_flavor: The telescope, should be LAT or SAT.

        wafer_slot: The wafer slot to get parameters for.

        config_path: Path to the yaml with the parameters.

    Returns:

        transform_pars: Dict of transformation parameters that can be passed to ufm_to_fp.
    """
    config = load_ufm_to_fp_config(config_path)
    return config[telescope_flavor][wafer_slot]


def ufm_to_fp(aman, x=None, y=None, pol=None, theta=0, dx=0, dy=0):
    """
    Transform from coords internal to wafer to focal plane coordinates.

    Arguments:

        aman: AxisManager assumed to contain aman.det_info.wafer.
              If provided outputs will be wrapped into aman.focal_plane.

        x: X position in wafer's internal coordinate system in mm.
           If provided overrides the value from aman.

        y: Y position in wafer's internal coordinate system in mm.
           If provided overrides the value from aman.

        pol: Polarization angle in wafer's internal coordinate system in deg.
           If provided overrides the value from aman.

        theta: Internal rotation of the UFM in degrees.

        dx: X offset in mm.

        dy: Y offset in mm.

    Returns:

        x_fp: X position on focal plane.

        y_fp: Y position on focal plane.

        pol_fp: Pol angle on focal plane.
    """
    if x is None:
        x = aman.det_info.wafer.det_x
    if y is None:
        y = aman.det_info.wafer.det_y
    if pol is None:
        pol = aman.det_info.wafer.angle
    xy = np.column_stack((x, y, np.zeros_like(x)))

    rot = R.from_euler("z", theta, degrees=True)
    xy = rot.apply(xy)

    x_fp = xy[:, 0] + dx
    y_fp = xy[:, 1] + dy
    pol_fp = pol + theta

    if aman is not None:
        focal_plane = core.AxisManager(aman.dets)
        focal_plane.wrap("x_fp", x_fp, [(0, focal_plane.dets)])
        focal_plane.wrap("y_fp", y_fp, [(0, focal_plane.dets)])
        focal_plane.wrap("pol_fp", pol_fp, [(0, focal_plane.dets)])
        if "focal_plane" in aman:
            aman.focal_plane.merge(focal_plane)
        else:
            aman.wrap("focal_plane", focal_plane)

    return x_fp, y_fp, pol_fp


def LAT_pix2sky(x, y, sec2el, sec2xel, array2secx, array2secy, rot=0, opt2cryo=30.0):
    """
    Routine to map pixels from arrays to sky.

    Arguments:

        x: X position on focal plane (currently zemax coord).

        y: Y position on focal plane (currently zemax coord).

        sec2el: Function that maps positions on secondary to on sky elevation. 

        sex2xel: Function that maps positions on secondary to on sky cross-elevation.

        array2secx: Function that maps positions on tube's focal plane to x position on secondary.

        array2secy: Function that maps positions on tube's focal plane to y position on secondary.

        rot: Rotation about the line of site = elev - 60 - corotator.

        opt2cryo: The rotation to get from cryostat coordinates to zemax coordinates (TBD, prob 30 deg).

    Returns:

        el: The on sky elevation in radians.

        xel: The on sky cross-elevation radians.
    """
    d2r = np.pi / 180.0
    # TBD - put in check for MASK - values outside circle should not be allowed
    # get into zemax coord
    xz = x * np.cos(d2r * opt2cryo) - y * np.sin(d2r * opt2cryo)
    yz = y * np.cos(d2r * opt2cryo) + x * np.sin(d2r * opt2cryo)
    # Where is it on (zemax secondary focal plane wrt LATR)
    xs = array2secx(xz, yz)
    ys = array2secy(xz, yz)
    # get into LAT zemax coord
    # There is an offset between how zemax coords and SO coords are defined
    rot += 180
    xrot = xs * np.cos(d2r * rot) - ys * np.sin(d2r * rot)
    yrot = ys * np.cos(d2r * rot) + xs * np.sin(d2r * rot)
    # note these are around the telescope boresight
    el = sec2el(xrot, yrot)
    xel = sec2xel(xrot, yrot)

    return np.deg2rad(el), np.deg2rad(xel)


@lru_cache(maxsize=None)
def load_zemax(path):
    """
    Load zemax_dat from path

    Arguments:

        path: Path to zemax data

    Returns:

        zemax_dat: Dictionairy with data from zemax
    """
    try:
        zemax_dat = np.load(path, allow_pickle=True)
    except Exception as e:
        logger.error("Can't load data from " + path)
        raise e
    return dict(zemax_dat)


@lru_cache(maxsize=None)
def LAT_optics(zemax_path):
    """
    Compute mapping from LAT secondary to sky.

    Arguments:

        zemax_path: Path to LAT optics data from zemax.

    Returns:

        sec2el: Function that maps positions on secondary to on sky elevation

        sex2xel: Function that maps positions on secondary to on sky cross-elevation.
    """
    zemax_dat = load_zemax(zemax_path)
    try:
        LAT = zemax_dat["LAT"][()]
    except Exception as e:
        logger.error("LAT key missing from dictionary")
        raise e

    gi = np.where(LAT["mask"] != 0.0)
    x = LAT["x"][gi].ravel()
    y = LAT["y"][gi].ravel()
    el = LAT["elev"][gi].ravel()
    xel = LAT["xel"][gi].ravel()

    s2e = bisplrep(x, y, el, kx=3, ky=3)
    sec2el = partial(_interp_func, spline=s2e)
    s2x = bisplrep(x, y, xel, kx=3, ky=3)
    sec2xel = partial(_interp_func, spline=s2x)

    return sec2el, sec2xel


@lru_cache(maxsize=None)
def LATR_optics(zemax_path, tube_slot):
    """
    Compute mapping from LAT secondary to sky.

    Arguments:

        zemax_path: Path to LATR optics data from zemax.

        tube_slot: Either the tube name as a string or the tube number as an int.

    Returns:

        array2secx: Function that maps positions on tube's focal plane to x position on secondary.

        array2secy: Function that maps positions on tube's focal plane to y position on secondary.
    """
    zemax_dat = load_zemax(zemax_path)
    try:
        LATR = zemax_dat["LATR"][()]
    except Exception as e:
        logger.error("LATR key missing from dictionary")
        raise e

    if isinstance(tube_slot, str):
        tube_name = tube_slot
        try:
            tube_num = LAT_TUBES[tube_slot]
        except Exception as e:
            logger.error("Invalid tube name")
            raise e
    elif isinstance(tube_slot, int):
        tube_num = tube_slot
        try:
            tube_name = list(LAT_TUBES.keys())[tube_num]
        except Exception as e:
            logger.error("Invalid tube number")
            raise e
    else:
        raise ValueError("Invalid tube_slot")

    logger.info("Working on LAT tube " + tube_name)
    gi = np.where(LATR[tube_num]["mask"] != 0)

    array_x = LATR[tube_num]["array_x"][gi].ravel()
    array_y = LATR[tube_num]["array_y"][gi].ravel()
    sec_x = LATR[tube_num]["sec_x"][gi].ravel()
    sec_y = LATR[tube_num]["sec_y"][gi].ravel()

    a2x = bisplrep(array_x, array_y, sec_x, kx=3, ky=3)
    array2secx = partial(_interp_func, spline=a2x)
    a2y = bisplrep(array_x, array_y, sec_y, kx=3, ky=3)
    array2secy = partial(_interp_func, spline=a2y)

    return array2secx, array2secy


def LAT_focal_plane(aman, zemax_path, x=None, y=None, pol=None, roll=0, tube_slot="c1"):
    """
    Compute focal plane for a wafer in the LAT.

    Arguments:

        aman: AxisManager nominally containing aman.focal_plane.x_fp and aman.focal_plane.y_fp.
              If provided focal plane will be stored in aman.focal_plane.

        zemax_path: Path to LATR optics data from zemax.

        x: Detector x positions, if provided will override positions loaded from aman.

        y: Detector y positions, if provided will override positions loaded from aman.

        pol: Detector polarization angle, if provided will override positions loaded from aman.

        roll: Rotation about the line of site = elev - 60 - corotator.

        tube_slot: Either the tube name as a string or the tube number as an int.

    Returns:

        xi: Detector xi on sky from physical optics in radians.
            If aman is provided then will be wrapped as aman.focal_plane.xi.

        eta: Detector eta on sky from physical optics in radians.
             If aman is provided then will be wrapped as aman.focal_plane.eta.

        gamma: Detector gamma on sky from physical optics in radians.
               If aman is provided then will be wrapped as aman.focal_plane.eta.
    """
    if x is None:
        x = aman.focal_plane.x_fp
    if y is None:
        y = aman.focal_plane.y_fp
    if pol is None:
        pol = aman.focal_plane.pol_fp

    sec2el, sec2xel = LAT_optics(zemax_path)
    array2secx, array2secy = LATR_optics(zemax_path, tube_slot)

    el, xel = LAT_pix2sky(x, y, sec2el, sec2xel, array2secx, array2secy, roll)
    xi, eta, _ = quat.decompose_xieta(quat.euler(1, np.deg2rad(90)) * quat.rotation_lonlat(-xel, el))

    pol_x, pol_y = gen_pol_endpoints(x, y, pol)
    pol_el, pol_xel = LAT_pix2sky(
        pol_x, pol_y, sec2el, sec2xel, array2secx, array2secy, roll
    )
    pol_xi, pol_eta, _ = quat.decompose_xieta(quat.euler(1, np.deg2rad(90)) * quat.rotation_lonlat(-pol_xel, pol_el))
    gamma = get_gamma(pol_xi, pol_eta)
    if np.isscalar(xi):
        gamma = gamma[0]

    if aman is not None:
        focal_plane = core.AxisManager(aman.dets)
        focal_plane.wrap("xi", xi, [(0, focal_plane.dets)])
        focal_plane.wrap("eta", eta, [(0, focal_plane.dets)])
        focal_plane.wrap("gamma", gamma, [(0, focal_plane.dets)])
        if "focal_plane" in aman:
            aman.focal_plane.merge(focal_plane)
        else:
            aman.wrap("focal_plane", focal_plane)

    return xi, eta, gamma


@lru_cache(maxsize=None)
def sat_to_sky(r_fp, r_sky):
    """
    Interpolate focal plane and on sky radius values to create mapping from SAT focal plane to sky.
    This function is a wrapper whose main purpose is the cache this mapping.

    Arguments:
        r_fp: Focal plane radius values in mm, should be all positive.

        theta: On sky radius values in radians, should be all positive.
               Theta is defined by ISO coordinates.

    Return:
        sat_to_sky: Interp object with the mapping from the focal plane to sky.
    """
    return make_interp_spline(r_fp, r_sky)


def SAT_focal_plane(aman, x=None, y=None, pol=None, roll=0, mapping_data=None):
    """
    Compute focal plane for a wafer in the SAT.

    Arguments:

        aman: AxisManager nominally containing aman.focal_plane.x_fp and aman.focal_plane.y_fp.
              If provided focal plane will be stored in aman.focal_plane.

        x: Detector x positions, if provided will override positions loaded from aman.

        y: Detector y positions, if provided will override positions loaded from aman.

        pol: Detector polarization angle, if provided will override positions loaded from aman.

        roll: Rotation about the line of site = -1*boresight.

        mapping_data: Tuple of (r_fp, r_sky) that can be interpolated to map the focal plane to the sky.
                      Leave as None to use the default mapping.

    Returns:

        xi: Detector xi on sky from physical optics in radians.
            If aman is provided then will be wrapped as aman.focal_plane.xi.

        eta: Detector eta on sky from physical optics in radians.
             If aman is provided then will be wrapped as aman.focal_plane.eta.

        gamma: Detector gamma on sky from physical optics in radians.
               If aman is provided then will be wrapped as aman.focal_plane.eta.
    """
    if x is None:
        x = aman.focal_plane.x_fp
    if y is None:
        y = aman.focal_plane.y_fp
    if pol is None:
        pol = aman.focal_plane.pol_fp

    if mapping_data is None:
        fp_to_sky = sat_to_sky(SAT_R_FP, SAT_R_SKY)
    else:
        mapping_data = (tuple(val) for val in mapping_data)
        fp_to_sky = sat_to_sky(*mapping_data)

    # Get things in polor coords 
    r_fp = (x**2 + y**2)**.5
    phi_fp = np.arctan2(y, x)

    # Now to the sky
    theta_iso = -fp_to_sky(r_fp)
    phi_iso = np.pi/2 - phi_fp

    # Flip about the origin for optics (180 deg rotation)
    phi_iso += np.pi

    # Now go to xieta
    _xi, _eta, _ = quat.decompose_xieta(quat.rotation_iso(theta_iso, phi_iso))

    # Apply roll
    xi = _xi * np.cos(np.deg2rad(roll)) - _eta * np.sin(np.deg2rad(roll))
    eta = _eta * np.cos(np.deg2rad(roll)) + _xi * np.sin(np.deg2rad(roll))

    # Lets do gamma as a set of endpoints
    pol_x, pol_y = gen_pol_endpoints(x, y, pol)
    pol_r = (pol_x**2 + pol_y**2)**.5
    pol_phi = np.arctan2(pol_y, pol_x)
    pol_theta_iso = -fp_to_sky(pol_r)
    pol_phi_iso = np.pi/2 - pol_phi + np.pi
    _xi, _eta, _ = quat.decompose_xieta(quat.rotation_iso(pol_theta_iso, pol_phi_iso))
    pol_xi = _xi * np.cos(np.deg2rad(roll)) - _eta * np.sin(np.deg2rad(roll))
    pol_eta = _eta * np.cos(np.deg2rad(roll)) + _xi * np.sin(np.deg2rad(roll))
    gamma = get_gamma(pol_xi, pol_eta)
    if np.isscalar(xi):
        gamma = gamma[0]

    if aman is not None:
        focal_plane = core.AxisManager(aman.dets)
        focal_plane.wrap("xi", xi, [(0, focal_plane.dets)])
        focal_plane.wrap("eta", eta, [(0, focal_plane.dets)])
        focal_plane.wrap("gamma", gamma, [(0, focal_plane.dets)])
        if "focal_plane" in aman:
            aman.focal_plane.merge(focal_plane)
        else:
            aman.wrap("focal_plane", focal_plane)

    return xi, eta, gamma


def get_focal_plane(
    aman,
    x=None,
    y=None,
    pol=None,
    roll=0,
    telescope_flavor=None,
    tube_slot=None,
    wafer_slot="ws0",
    config_path=None,
    ufm_to_fp_pars=None,
    zemax_path=None,
    mapping_data=None,
    return_fp=False,
):
    """
    Unified interface for LAT_focal_plane and SAT_focal_plane including the step of applying ufm_to_fp.

    Arguments:

        aman: AxisManager nominally containing aman.det_info.wafer.det_x and aman.det_info.wafer.det_y.
              If provided focal plane will be stored in aman.focal_plane.

        x: Detector x positions, if provided will override positions loaded from aman.

        y: Detector y positions, if provided will override positions loaded from aman.

        pol: Detector polarization angle, if provided will override positions loaded from aman.

        roll: Rotation about the line of sight.
              For the LAT this is elev - 60 - corotator.
              For the SAT this is -1*boresight.

        telescope_flavor: What the telescope flavor is ('LAT' or 'SAT').
                          Note that this is case insenstive.
                          If None and aman contains obs_info, will be extracted from there.

        tube_slot: Which optics tube of the telescope to use.
                   Only used by the LAT.
                   If None and aman contains obs_info, will be extracted from there.

        wafer_slot: Which wafer slot to use.

        config_path: Path to the ufm_to_fp config file.

        ufm_to_fp_pars: Loaded ufm_to_fp_parsm to focalplane params.
                        If provided config_path is is ignored.

        zemax_path: Path to the data file from Zemax.
                    Only used by the LAT.

        mapping_data: Tuple of (x, theta) that can be interpolated to map the focal plane to the sky.
                      Leave as None to use the default mapping.
                      Only used by the SAT.

        return_fp: If True also return the transformed focal plane in mm.

    Returns:

        xi: Detector xi on sky from physical optics in radians.
            If aman is provided then will be wrapped as aman.focal_plane.xi.

        eta: Detector eta on sky from physical optics in radians.
             If aman is provided then will be wrapped as aman.focal_plane.eta.

        gamma: Detector gamma on sky from physical optics in radians.
               If aman is provided then will be wrapped as aman.focal_plane.eta.

        x_fp: X position on focal plane.
              Only returned if return_fp is True.

        y_fp: Y position on focal plane.
              Only returned if return_fp is True.

        pol_fp: Pol angle on focal plane.
                Only returned if return_fp is True.
    """
    if aman is not None and "obs_info" in aman:
        if telescope_flavor is None:
            telescope_flavor = aman.obs_info.telescope_flavor
        if tube_slot is None:
            tube_slot = aman.obs_info.tube_slot
    if telescope_flavor is None or tube_slot is None:
        raise ValueError("Telescope or tube is None and unable to extract from aman")
    telescope_flavor = telescope_flavor.upper()
    if telescope_flavor not in ["LAT", "SAT"]:
        raise ValueError("Telescope should be LAT or SAT")

    if ufm_to_fp_pars is None:
        ufm_to_fp_pars = get_ufm_to_fp_pars(telescope_flavor, wafer_slot, config_path)
    x_fp, y_fp, pol_fp = ufm_to_fp(aman, x=x, y=y, pol=pol, **ufm_to_fp_pars)
    if telescope_flavor == "LAT":
        if zemax_path is None:
            raise ValueError("Must provide zemax_path for LAT")
        xi, eta, gamma = LAT_focal_plane(
            aman, zemax_path, x=x_fp, y=y_fp, pol=pol_fp, roll=roll, tube_slot=tube_slot
        )
    else:
        xi, eta, gamma = SAT_focal_plane(
            aman, x=x_fp, y=y_fp, pol=pol_fp, roll=roll, mapping_data=mapping_data
        )

    if return_fp:
        return xi, eta, gamma, x_fp, y_fp, pol_fp
    return xi, eta, gamma


def gen_template(wafer_info, stream_id, **pointing_cfg):
    """
    Generate a pointing template from a wafer info ResultSet.

    Arguments:

        wafer_info: Either the path to a wafer_info ResultSet
                    or an open h5py File object.

        stream_id : The UFM to generate the template for (ie: ufm_mv5).

        **pointing_cfg: Arguments to pass to get_focal_plane,
                        should be arguments from rot onwards.

    Returns:

        template_det_ids: The detector ids of the detectors in the template.

        template: (ndet, 3) array of (xi, eta, gamma) for all dets in template.

        is_optical: (ndet) mask that is True for all optical dets.
    """
    ufm = stream_id.split("_")[-1].title()
    wafer = read_dataset(wafer_info, ufm)
    template_det_ids = wafer["dets:det_id"]
    det_x = wafer["dets:wafer.x"]
    det_y = wafer["dets:wafer.y"]
    det_pol = wafer["dets:wafer.angle"]
    det_type = wafer["dets:wafer.type"]
    is_optical = np.array(det_type) == "OPTC"

    xi, eta, gamma = get_focal_plane(
        None, x=det_x, y=det_y, pol=det_pol, **pointing_cfg
    )
    template = np.column_stack((xi, eta, gamma))

    # Dark dets should not be allowed to have pointing
    template[~is_optical] = np.nan

    return np.array(template_det_ids), template, is_optical



# Copyright (c) 2022-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

import os
import re
import datetime

import numpy as np
import traitlets
from astropy import units as u
import yaml

from toast.timing import function_timer, Timer
from toast.traits import trait_docs, Int, Unicode, Bool, Quantity, Float, Instance, List
from toast.ops.operator import Operator
from toast.utils import Environment, Logger
from toast.dist import distribute_discrete
from toast.observation import default_values as defaults

from ..io import parse_book_name, parse_book_time, read_book
from ...sim_hardware import sim_nominal
from ..instrument import SOFocalplane


@trait_docs
class LoadBooks(Operator):
    """Load Level-3 Books into observations.

    If no focalplane directory is specified, a minimal focalplane is created using only
    the stream_id and readout names in the G3 files.  Similarly for noise models.  If
    the noise_dir is specified, noise models are loaded from there.

    """

    # Class traits

    API = Int(0, help="Internal interface version for this operator")

    books = List(list(), help="List of observation book directories")

    focalplane_dir = Unicode(
        None, allow_none=True, help="Directory for focalplane models"
    )

    detset_key = Unicode(
        None,
        allow_none=True,
        help="If specified, use this column of the focalplane detector_data to group detectors",
    )

    noise_dir = Unicode(None, allow_none=True, help="Directory for noise models")

    wafers = List(list(), help="Only load detectors from these wafers / stream_ids")

    bands = List(list(), help="Only load detectors from these bands")

    times = Unicode(defaults.times, help="Observation shared key for timestamps")

    shared_flags = Unicode(
        defaults.shared_flags,
        allow_none=True,
        help="Observation shared key for common flags",
    )

    det_data = Unicode(
        defaults.det_data,
        allow_none=True,
        help="Observation detdata key for detector signal",
    )

    det_flags = Unicode(
        defaults.det_flags,
        allow_none=True,
        help="Observation detdata key for detector flags",
    )

    azimuth = Unicode(
        defaults.azimuth, help="Observation shared key for boresight Azimuth"
    )

    elevation = Unicode(
        defaults.elevation, help="Observation shared key for boresight Elevation"
    )

    boresight_azel = Unicode(
        defaults.boresight_azel,
        help="Observation shared key for boresight Az/El quaternions",
    )

    boresight_radec = Unicode(
        defaults.boresight_radec,
        help="Observation shared key for boresight RA/DEC quaternions",
    )

    corotator_angle = Unicode(
        defaults.corotator_angle,
        allow_none=True,
        help="Observation shared key for corotator_angle (if it is used)",
    )

    boresight_angle = Unicode(
        defaults.boresight_angle,
        allow_none=True,
        help="Observation shared key for boresight rotation angle (if it is used)",
    )

    hwp_angle = Unicode(
        defaults.hwp_angle,
        allow_none=True,
        help="Observation shared key for HWP angle",
    )

    frame_intervals = Unicode(
        None,
        allow_none=True,
        help="Observation interval key for frame boundaries",
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @function_timer
    def _exec(self, data, detectors=None, **kwargs):
        log = Logger.get()
        timer = Timer()
        timer.start()

        if self.books is None or len(self.books) == 0:
            raise RuntimeError("No books specified, nothing to load")

        # Extract the frequency "name" from any specified bands
        band_freqs = list()
        band_pat = re.compile(r".*(\d+)")
        if self.bands is not None:
            for bnd in self.bands:
                mat = band_pat.match(bnd)
                if mat is None:
                    raise RuntimeError(
                        "If specified, band names must end with the frequency"
                    )
                band_freqs.append(int(mat.group(1)))

        book_props = None
        if data.comm.group_rank == 0:
            # We are only using the tube / wafer / band information here
            # (not detector properties) so the nominal hardware model should be fine.
            hw = sim_nominal()

            # Parse the book names and corresponding information
            book_pat = re.compile(r"obs_(\d+)_(.*)_(\d+)")
            book_props = dict()
            for path in self.books:
                book_name = os.path.basename(path)
                book_timestamp, book_tele, book_tube, book_wafers = parse_book_name(
                    book_name
                )
                # print(
                #     f"Load found book {book_timestamp}, {book_tele}, {book_tube}, {book_wafers}"
                # )
                wf_props = dict()
                for wf in book_wafers:
                    if len(self.wafers) > 0 and wf not in self.wafers:
                        # print(f"  wafer {wf} not in list, skipping")
                        continue
                    bands = hw.data["wafer_slots"][wf]["bands"]
                    npix = hw.data["wafer_slots"][wf]["npixel"]
                    use_bands = list()
                    for bnd in bands:
                        bmat = band_pat.match(bnd)
                        # if bmat is None:
                        #     print(f"  band {bnd} does not match- should never happen!")
                        bfreq = int(bmat.group(1))
                        if len(band_freqs) > 0 and bfreq not in band_freqs:
                            # print(f"  band {bnd} ({bfreq}) not in list, skipping")
                            continue
                        use_bands.append(bnd)
                    if len(use_bands) > 0:
                        # print(f"  Adding wafer {wf}")
                        wf_props[wf] = {
                            "npixel": npix,
                            "bands": use_bands,
                        }
                    # else:
                    #     print(f"skipping wafer {wf} with no selected bands")
                if len(wf_props) == 0:
                    msg = (
                        f"book '{book_name}' had no wafers matching wafer / band lists"
                    )
                    log.warning(msg)
                else:
                    # Use this book.
                    book_props[book_name] = {
                        "path": path,
                        "timestamp": book_timestamp,
                        "telescope": book_tele,
                        "tube": book_tube,
                        "wafers": wf_props,
                    }
        if data.comm.comm_group is not None:
            book_props = data.comm.comm_group.bcast(book_props, root=0)

        if len(book_props) == 0:
            raise RuntimeError("No books found matching wafer / band criteria")

        # Determine the approximate size of each book, for load balancing among groups

        book_names = list(sorted(book_props.keys()))
        book_sizes = list()
        if data.comm.group_rank == 0:
            for name in book_names:
                props = book_props[name]
                m_obs_file = os.path.join(props["path"], "M_index.yaml")
                if not os.path.isfile(m_obs_file):
                    msg = f"Book directory {path} does not contain 'M_index.yaml'"
                    log.error(msg)
                    raise RuntimeError(msg)
                with open(m_obs_file, "r") as f:
                    m_obs = yaml.load(f, Loader=yaml.FullLoader)
                start = parse_book_time(m_obs["timestamp_start"])
                end = parse_book_time(m_obs["timestamp_end"])
                duration = (end - start).total_seconds() / 3600.0
                total_pix = np.sum(
                    [
                        (v["npixel"] * len(v["bands"]))
                        for k, v in props["wafers"].items()
                    ]
                )
                book_sizes.append(duration * total_pix)
        if data.comm.comm_group is not None:
            book_sizes = data.comm.comm_group.bcast(book_sizes, root=0)

        # Dictionary of observation field names
        obs_fields = {
            "times": self.times,
            "shared_flags": self.shared_flags,
            "det_data": self.det_data,
            "det_flags": self.det_flags,
            "hwp_angle": self.hwp_angle,
            "azimuth": self.azimuth,
            "elevation": self.elevation,
            "boresight_azel": self.boresight_azel,
            "boresight_radec": self.boresight_radec,
            "corotator_angle": self.corotator_angle,
            "boresight_angle": self.boresight_angle,
        }

        # Distribute books among groups.

        groupdist = distribute_discrete(book_sizes, data.comm.ngroups)

        # Every process group creates observations from their books

        group_firstbook = groupdist[data.comm.group][0]
        group_numbooks = groupdist[data.comm.group][1]

        fp_cache = dict()

        for book_indx in range(group_firstbook, group_firstbook + group_numbooks):
            bname = book_names[book_indx]
            bprops = book_props[bname]
            telename = bprops["telescope"]

            msg = f"Process group {data.comm.group} loading book {bname}"
            log.info_rank(msg, comm=data.comm.comm_group)
            if telename in fp_cache:
                nominal_fp = fp_cache[telename]
            else:
                # We are only using this for a nominal channel map- the
                # sample rate does not matter.
                nominal_fp = SOFocalplane(
                    sample_rate=1.0 * u.Hz,
                    telescope=telename,
                )
                fp_cache[telename] = nominal_fp

            session_obs = read_book(
                data.comm,
                nominal_fp,
                bprops["path"],
                self.focalplane_dir,
                self.noise_dir,
                obs_fields,
                bprops["tube"],
                bprops["wafers"],
                detset_key=self.detset_key,
                frame_intervals=self.frame_intervals,
            )

            # Redistribute each observation and append to the output.  Also
            # create a set of intervals corresponding to the frame boundaries.
            for ob in session_obs:
                ob.redistribute(data.comm.group_size, times=obs_fields["times"])
                data.obs.append(ob)

    def _finalize(self, data, **kwargs):
        return

    def _requires(self):
        return dict()

    def _provides(self):
        prov = {
            "shared": [
                self.times,
                self.boresight_azel,
            ],
            "detdata": [self.det_data],
        }
        if self.boresight_radec is not None:
            prov["shared"].append(self.boresight_radec)
        if self.boresight_angle is not None:
            prov["shared"].append(self.boresight_angle)
        if self.corotator_angle is not None:
            prov["shared"].append(self.corotator_angle)
        if self.hwp_angle is not None:
            prov["shared"].append(self.hwp_angle)
        return prov



# Copyright (c) 2018-2019 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Test hardware model creation, dump, and load.
"""

import os

import unittest
from unittest import TestCase
from collections import OrderedDict

from sotodlib.core import Hardware
from sotodlib.sim_hardware import sim_nominal
from sotodlib.vis_hardware import plot_detectors

from ._helpers import create_outdir, mpi_multi

try:
    # Import sotodlib toast module first, which sets global toast defaults
    import sotodlib.toast as sotoast
    import toast
    from sotodlib.toast import sim_focalplane as toastsf

    toast_available = True
except ImportError:
    toast_available = False


@unittest.skipIf(mpi_multi(), "Running with multiple MPI processes")
class HardwareTest(TestCase):

    def setUp(self):
        fixture_name = os.path.splitext(os.path.basename(__file__))[0]
        self.outdir = create_outdir(fixture_name)
        self.skip_plots = False
        if "SOTODLIB_TEST_DISABLE_PLOTS" in os.environ:
            self.skip_plots = os.environ["SOTODLIB_TEST_DISABLE_PLOTS"]

    def test_config_example(self):
        outpath = os.path.join(self.outdir, "hardware_example.toml.gz")
        hw = sim_nominal()
        hw.dump(outpath, overwrite=True, compress=True)
        hwcheck = Hardware(outpath)
        checkpath = os.path.join(self.outdir, "hardware_example_check.toml.gz")
        hwcheck.dump(checkpath, overwrite=True, compress=True)
        return

    def test_sim_wafer(self):
        if not toast_available:
            print("TOAST cannot be imported- skipping unit test", flush=True)
            return
        hw = sim_nominal()
        # Simulate some wafers
        for tele, teleprops in hw.data["telescopes"].items():
            if tele != "SAT1":
                continue
            platescale = teleprops["platescale"]
            fwhm = teleprops["fwhm"]
            for tube_slot in teleprops["tube_slots"]:
                if tube_slot != "ST1":
                    continue
                tubeprops = hw.data["tube_slots"][tube_slot]
                for wafer in tubeprops["wafer_slots"]:
                    if wafer != "w25":
                        continue
                    outpath = os.path.join(
                        self.outdir, "wafer_{}.toml.gz".format(wafer))
                    del hw.data["detectors"]
                    dets = toastsf.sim_wafer_detectors(hw, wafer, platescale, fwhm)
                    hw.data["detectors"] = dets
                    hw.dump(outpath, overwrite=True, compress=True)
                    if not self.skip_plots:
                        outpath = os.path.join(
                            self.outdir,
                            "wafer_{}.pdf".format(wafer)
                        )
                        plot_detectors(hw.data["detectors"], outpath, labels=True)
        return

    def test_sim_telescope(self):
        if not toast_available:
            print("TOAST cannot be imported- skipping unit test", flush=True)
            return
        fullhw = sim_nominal()

        # SAT test
        toastsf.sim_telescope_detectors(fullhw, "SAT1")
        hw = fullhw.select(match={"wafer_slot": ["w25",],})
        outpath = os.path.join(self.outdir, "telescope_SAT1_w25.toml.gz")
        hw.dump(outpath, overwrite=True, compress=True)
        if not self.skip_plots:
            outpath = os.path.join(self.outdir, "telescope_SAT1_w25.pdf")
            plot_detectors(hw.data["detectors"], outpath, labels=False)

        fullhw.data["detectors"] = OrderedDict()

        # LAT test
        toastsf.sim_telescope_detectors(fullhw, "LAT")
        hw = fullhw.select(match={"tube_slots": ["c1",],})
        outpath = os.path.join(self.outdir, "telescope_LAT_c1.toml.gz")
        hw.dump(outpath, overwrite=True, compress=True)
        if not self.skip_plots:
            outpath = os.path.join(self.outdir, "telescope_LAT_c1.pdf")
            plot_detectors(hw.data["detectors"], outpath, labels=False)

        return

    def test_sim_full(self):
        if not toast_available:
            print("TOAST cannot be imported- skipping unit test", flush=True)
            return
        hw = sim_nominal()
        for tele, teleprops in hw.data["telescopes"].items():
            if tele != "SAT1":
                continue
            toastsf.sim_telescope_detectors(hw, tele)
        dbpath = os.path.join(self.outdir, "hardware_SAT1.toml.gz")
        hw.dump(dbpath, overwrite=True, compress=True)
        check = Hardware(dbpath)

        # Test selection of 90GHz detectors on wafers 25 and 26 which have
        # "A" polarization configuration and are located in pixels 20-29.
        wbhw = hw.select(
            match={"wafer_slot": ["w25", "w26"],
                "band": "SAT_f090",
                "pol": "A",
                "pixel": "02."})
        dbpath = os.path.join(self.outdir, "w25-26_p20-29_SAT_f090_A.toml.gz")
        wbhw.dump(dbpath, overwrite=True, compress=True)
        check = Hardware(dbpath)
        self.assertTrue(len(check.data["detectors"]) == 20)
        chkpath = os.path.join(self.outdir, "w25-26_p20-29_SAT_f090_A.txt")
        with open(chkpath, "w") as f:
            for d in check.data["detectors"]:
                f.write("{}\n".format(d))
        return



"""check_book.py

This module contains a class for checking obs/oper Books (for internal
consistency, proper schema, etc).  The results can also be used to
help update an ObsFileDb.

"""

import so3g
from spt3g import core

import os
import fnmatch
import glob
import logging
import yaml

import numpy as np


logger = logging.getLogger(__name__)

# The default config for BookScanner.

_DEFAULT_CONFIG = {
    # File patterns ...
    'stream_file_pattern': 'D_{stream_id}_{index:03d}.g3',
    'ancil_file_pattern': 'A_ancil_{index:03d}.g3',

    # Schema ...
    'extra_files': ['M_index.yaml', 'M_book.yaml'],
    'extra_extra_files': [],
    'banned_files': [],

    # Work-arounds ...
    'sample_range_inclusive_hack': False,
    'require_ordered_detsets': True,
    'tolerate_sample_range_resetting': False,
    'tolerate_missing_ancil': False,
    'tolerate_missing_ancil_timestamps': False,
    'tolerate_timestamps_value_discrepancy': False,
    'tolerate_stray_files': False,
    'tolerate_missing_extra_files': False,
    'ignore_fixed_tones': False,
}
    

class _compact_list(list):
    # list with abbreviated repr.
    def __repr__(self):
        others = len(self) - 5
        if len(self) > 10:
            return '[' + ', '.join([repr(x) for x in self[:5]]) \
                + f', and {others} other items ...]'
        else:
            return super.__repr__(self)


class BookScanner:
    """The BookScanner helps to catalog the contents of an obs/oper book,
    validate that the contents look right, and produce entries for
    ObsFileDb.

    It is used like this::

      # Instantiate with path to a obs/oper Book.
      bs = BookScanner(book_dir)

      # Analyze book contents.
      bs.go()

      # Optional: print summary of warnings / errors to terminal.
      bs.report()

      # Check for errors.
      if len(bs.results['errors']):
          raise RuntimeError("This book contains errors.")

      # Get stuff for the obsfiledb (paths relative to ...)
      detset_rows, file_rows = bs.get_obsfiledb_info('/')

    """
    logger = logger  # cache module-level logger

    def __init__(self, book_dir, config={}, logger=None):
        self.config = dict(_DEFAULT_CONFIG, **config)
        self.book_dir = book_dir
        self.results = {
            'errors': [],
            'warnings': [],
            'det_lists': {},
            'sample_ranges': None,
            'ready': False,
            'metadata': None,
        }
        if logger is not None:
            self.logger = logger

    def go(self):
        """Scan a book for completeness and consistency; record misc. metadata
        to help with ObsFileDb registration.

        """
        # 1. Check metadata completeness.
        try:
            self.check_meta()
            self.gapfill_meta()
        except Exception as e:
            self.logger.error('Failed in section 1!')
            self.report()
            raise e

        # 2. Verify file presence
        try:
            self.check_file_presence()
        except Exception as e:
            self.logger.error('Failed in section 2!')
            self.report()
            raise e

        try:
            # 3. Check sample range, timestamp alignment, etc.
            self.check_frames()
        except Exception as e:
            self.logger.error('Failed in section 3!')
            self.report()
            raise e

    def _err(self, filename, msg, warn=False):
        if warn:
            self.results['warnings'].append((filename, msg))
        else:
            self.results['errors'].append((filename, msg))

    def _get_filename(self, basename, **kw):
        basename = basename.format(**kw)
        return os.path.join(self.book_dir, basename)

    def check_meta(self):
        """Extract essential metadata from the M_index.yaml.

        """
        err, warn = self.results['errors'], self.results['warnings']
        meta = yaml.safe_load(open(self._get_filename('M_index.yaml'), 'rb'))
        self.results['metadata'] = meta

        keys = ['book_id', 'type', 'telescope', 'stream_ids', 'sample_ranges']
        for k in keys:
            if k not in meta:
                self._err('M_index.yaml', f'Missing entry for "{k}".', warn=True)

    def gapfill_meta(self):
        """If some metadata are missing, try to guess them from the filenames
        and stuff.

        """
        meta = self.results['metadata']

        if 'book_id' not in meta:
            meta['book_id'] = os.path.split(self.book_dir)[1]
        tokens = meta['book_id'].split('_')  # type, time, tel, slot-flags
        if 'type' not in meta:
            meta['type'] = tokens[0]
        if 'telescope' not in meta:
            meta['telescope'] = tokens[2]
        if 'stream_ids' not in meta:
            assert 'stream_ids' in self.config  # needed to patch missing stream_id info
            meta['stream_ids'] = []
            for sid, f in zip(self.config['stream_ids'], tokens[3]):
                if f == '0':
                    continue
                elif f == '1':
                    meta['stream_ids'].append(sid)
                else:
                    self._err(
                        None, f'Invalid stream_flags in book name: {tokens[3]}.')

        if 'detsets' not in meta:
            assert 'detset_map' in self.config
            meta['detsets'] = [self.config['detset_map'][k]
                               for k in meta['stream_ids']]

        # Check that detsets and stream_ids are same length.
        assert len(meta['detsets']) == len(meta['stream_ids'])

        # Don't check sample_ranges here, do it later when we're
        # reading the files for per-frame sample ranges.

    def check_file_presence(self):
        """See if all expected files are there; compare number of files in
        each file group.

        """
        strays = glob.glob(self._get_filename('*'))
        for basename in self.config['extra_files'] + self.config['extra_extra_files']:
            fn = self._get_filename(basename)
            try:
                strays.remove(fn)
            except ValueError:
                self._err(basename, 'Expected "extra file" and did not find it.',
                          warn=self.config['tolerate_missing_extra_files'])

        meta = self.results['metadata']
        file_counts = []

        for stream_id in meta['stream_ids'] + ['ancil']:
            if stream_id == 'ancil':
                pattern = self.config['ancil_file_pattern']
            else:
                pattern = self.config['stream_file_pattern']
            index = 0
            while True:
                fn = self._get_filename(pattern, stream_id=stream_id, index=index)
                if not os.path.exists(fn):
                    break
                strays.remove(fn)
                index += 1
            file_counts.append(index)

        file_count = file_counts[0]
        if any([f != file_count for f in file_counts]):
            self._err(None, f'Inconsistent file counts: {file_counts}')

        _file_count = file_count if 'sample_ranges' not in meta \
                      else len(meta['sample_ranges'])
        if _file_count != file_count:
            self._err(None, f'File counts {file_counts} do not agree with '
                      'sample_ranges entry {_file_count}')
        meta['file_count'] = file_count
        if len(strays):
            self._err(None, f'Extra files found in book dir: {strays}',
                      warn=self.config['tolerate_stray_files'])
            for bf in self.config['banned_files']:
                # interpret as a pattern
                for banned in fnmatch.filter(strays, self._get_filename(bf)):
                    self._err(banned, f'File is banned from book by pattern: {bf}')

    def check_frames(self):
        """Read the "Scan" frames from all files and check structure /
        consistency.

        """
        meta = self.results['metadata']
        to_check = ['ancil'] + meta['stream_ids']

        timestamps_master = None
        sample_ranges = []
        for stream_id in to_check:
            if stream_id == 'ancil':
                pattern = self.config['ancil_file_pattern']
            else:
                pattern = self.config['stream_file_pattern']

            timestamps = []
            start, end = 0, None
            hack_offset = 0
            dets = None

            for index in range(meta['file_count']):
                filename = self._get_filename(pattern, stream_id=stream_id, index=index)
                basename = os.path.split(filename)[1]
                for frame in core.G3File(filename):
                    if frame.type == core.G3FrameType.Scan:
                        a, b = list(frame['sample_range'])
                        if self.config['sample_range_inclusive_hack']:
                            # Change (0, 999) into (0, 1000).
                            b = b + 1  # schema hack

                        if end is None:
                            # This is first frame in new file and
                            # "start" is expected first sample index.
                            if a == 0 and start != 0:
                                self._err(basename, 'sample_range is resetting on file boundaries.',
                                          warn=self.config['tolerate_sample_range_resetting'])
                                hack_offset = start
                            if a != start - hack_offset:
                                self._err(basename, f'sample_range entries are not abutting: '
                                          f'[..., {start - hack_offset}] -> [{a}, {b}].')
                                raise RuntimeError()
                        else:
                            # This is not the first frame, "end" from
                            # last frame should match start of this one.
                            if a != end - hack_offset:
                                self._err(basename, f'sample_range entries are not abutting: '
                                          f'[..., {end - hack_offset}] -> [{a}, {b}].')
                                raise RuntimeError()
                        end = b + hack_offset

                        if 'ancil' in frame:
                            timestamps.append(np.asarray(frame['ancil'].times))
                        else:
                            self._err(basename, f'No "ancil" entry',
                                      warn=self.config['tolerate_missing_ancil'])
                            if 'primary' in frame:
                                timestamps.append(np.asarray(frame['primary'].times))

                        if 'signal' in frame and dets is None:
                            dets = _compact_list(frame['signal'].names)

                # Wrap up this file
                if index >= len(sample_ranges):
                    sample_ranges.append((start, end))
                else:
                    if sample_ranges[index] != (start, end):
                        self._err(basename, f'sample_range discrepancy, {(start, end)} '
                                  f'instead of {sample_ranges[index]}')

                # For next file, requeue expectations.
                start, end = end, None

            # Check / record stuff for the stream_id.
            self.results['det_lists'][stream_id] = dets

            if len(timestamps) == 0:
                self._err(None, f'No timestamps in fileset.',
                          warn=(stream_id == 'ancil' and self.config['tolerate_missing_ancil_timestamps']))
            else:
                timestamps = np.hstack(timestamps)
                if timestamps_master is None:
                    timestamps_master = timestamps
                else:
                    if len(timestamps) != len(timestamps_master):
                        self._err(None, f'Timestamps length discrepancy.')
                    elif not (timestamps == timestamps_master).all():
                        dmax_us = np.max(abs(timestamps - timestamps_master)) \
                                  / core.G3Units.microseconds
                        self._err(None, f'Timestamps value discrepancy (up to {dmax_us:.3} us).',
                                  warn=self.config['tolerate_timestamps_value_discrepancy'])

        # If metadata included sample_ranges, check that.
        _sample_ranges = meta.get('sample_ranges')
        if _sample_ranges is not None:
            A, B = np.array(_sample_ranges), np.array(sample_ranges)
            if A.shape != B.shape or np.any(A != B):
                self._err(None, f'sample_ranges from metadata is not as found in files.')
        self.results['sample_ranges'] = sample_ranges

    def get_obsfiledb_info(self, filebase_root):
        """Get entries ready for obsfiledb addition.

        Args:
          filebase_root (str): Path relative to which the file paths should
            be specified.  If this is '/' then an absolute path will
            be recorded.

        Returns:
          detset_rows: list of implicated detsets, with each entry a
            tuple (detset_name, detset_dets).
          file_rows: list of file table entries; each entry is a dict
            that can be passed as kwargs to ObsFileDb.add_file.

        """
        meta, det_lists, sample_ranges = [
            self.results[k] for k in ['metadata', 'det_lists', 'sample_ranges']]
        detset_rows = []
        file_rows = []

        for stream_id, detset in zip(meta['stream_ids'], meta['detsets']):
            if self.config['require_ordered_detsets']:
                assert stream_id in detset
            elif stream_id not in detset:
                # if detsets don't need to be in the correct order just find the
                # right one.
                detset = [x for x in meta['detsets'] if stream_id in x][0]

            # Screen out "fixed tones".  We don't want these in the
            # obsfiledb detsets.
            dets = [d for d in det_lists[stream_id] if "NONE" not in d]
            n_fixed_tone = len(det_lists[stream_id]) - len(dets)
            if not self.config['ignore_fixed_tones']:
                # Yes, "ignore" just means don't even talk about it.
                self.logger.warn(f'Suppressing {n_fixed_tone} fixed tones '
                                 f'in {stream_id}')
            detset_rows.append((detset, dets))

            for index in range(meta['file_count']):
                pattern = self.config['stream_file_pattern']
                path = self._get_filename(
                    pattern, stream_id=stream_id, index=index)
                if filebase_root == '/':
                    relpath = os.path.abspath(path)
                else:
                    relpath = os.path.relpath(path, filebase_root)
                file_rows.append(
                    {'filename': relpath,
                     'obs_id': meta['book_id'],
                     'detset': detset,
                     'sample_start': sample_ranges[index][0],
                     'sample_stop': sample_ranges[index][1],
                     })

        return detset_rows, file_rows

    def report(self):
        """Print a summary of warning and error messages."""
        self.logger.info('Warnings:')
        for err in self.results['warnings']:
            self.logger.info(err)
        self.logger.info('Errors:')
        for err in self.results['errors']:
            self.logger.info(err)




#!/usr/bin/env python

# Copyright (c) 2019 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

import argparse
import os
import sys
from collections import OrderedDict

import numpy as np
from numpy.random import randn

from ... import sim_hardware as hardware


def write_hw(out, hw, plain=False, overwrite=True):
    if plain:
        outpath = "{}.toml".format(out)
    else:
        outpath = "{}.toml.gz".format(out)
    print("Dumping config to {}...".format(outpath))
    hw.dump(outpath, overwrite=overwrite, compress=(not plain))
    return


def main():
    parser = argparse.ArgumentParser(
        description="This program simulates the current nominal hardware "
        "model, perturbs it and dumps it to disk",
        usage="make_hardware_maps [options] (use --help for details)")

    parser.add_argument(
        "--out", required=False, default="hardware",
        help="Name (without extensions) of the output hardware file")
    parser.add_argument(
        "--plain", required=False, default=False, action="store_true",
        help="Write plain text (without gzip compression)")

    parser.add_argument(
        "--sigma_band_width", default=0.01, type=np.float,
        help="Proportional width of top hat bandpass width distribution.")
    parser.add_argument(
        "--sigma_band_center", default=0.01, type=np.float,
        help="Proportional width of top hat bandpass center distribution.")
    parser.add_argument(
        "--sigma_noise_net", default=0.01, type=np.float,
        help="Proportional width of noise NET distribution.")
    parser.add_argument(
        "--sigma_noise_fknee", default=0.01, type=np.float,
        help="Proportional width of noise fknee distribution.")
    parser.add_argument(
        "--sigma_noise_fmin", default=0.01, type=np.float,
        help="Proportional width of noise fmin distribution.")
    parser.add_argument(
        "--sigma_noise_alpha", default=0.01, type=np.float,
        help="Proportional width of noise alpha distribution.")
    parser.add_argument(
        "--sigma_beam_fwhm", default=0.01, type=np.float,
        help="Proportional width of beam FWHM distribution.")

    args = parser.parse_args()

    # Nominal configuration

    print("Getting nominal config...", flush=True)
    hw = hardware.sim_nominal()
    for tele, teleprops in hw.data["telescopes"].items():
        print("Simulating detectors for telescope {}...".format(tele),
              flush=True)
        hardware.sim_detectors_toast(hw, tele)
    write_hw(args.out + '.nominal', hw, plain=args.plain)

    # Perturb configuration

    print("Perturbing nominal config...", flush=True)
    for detname, detdata in hw.data["detectors"].items():
        bandname = detdata["band"]
        banddata = hw.data["bands"][bandname]
        # Perturb bandpass
        low = banddata["low"]
        center = banddata["center"]
        high = banddata["high"]
        halfband = .5 * (high - low) * (1 + randn() * args.sigma_band_width)
        center *= 1 + randn() * args.sigma_band_center
        low = center - halfband
        high = center + halfband
        detdata["low"] = low
        detdata["center"] = center
        detdata["high"] = high
        # Perturb noise
        detdata["NET"] = banddata["NET"] * (1 + randn() * args.sigma_noise_net)
        detdata["fknee"] = banddata["fknee"] * (1 + randn() * args.sigma_noise_fknee)
        detdata["fmin"] = banddata["fmin"] * (1 + randn() * args.sigma_noise_fmin)
        detdata["alpha"] = banddata["alpha"] * (1 + randn() * args.sigma_noise_alpha)
        # Perturb beam
        detdata["fwhm"] *= (1 + randn() * args.sigma_beam_fwhm)
        # Perturb position?  This would be the place to add errors in focal plane geometry
    write_hw(args.out + '.perturbed', hw, plain=args.plain)

    return


if __name__ == "__main__":
    main()



from dataclasses import dataclass, fields
from typing import Literal, Any, Dict, Union, Optional, List, Tuple
import datetime as dt
import yaml
import pandas as pd
import requests
from tqdm import tqdm
from io import StringIO
import json
import h5py
import logging
import numpy as np
# import alphashape
from influxdb import InfluxDBClient
from collections import defaultdict

from sotodlib.io import hkdb
from sotodlib.io.hkdb import HkConfig
from sotodlib.core import Context
from sotodlib.core.metadata import ManifestDb

logger = logging.getLogger(__name__)

class ReportDataConfig:
    def __init__(
        self,
        platform: Literal["satp1", "satp2", "satp3", "lat"],
        site_url: str,
        ctx_path: str,
        start_time: Union[dt.datetime, float, str],
        stop_time: Union[dt.datetime, float, str],
        hk_cfg: Union[HkConfig, str, Dict[str, Any]],
        buffer_time: float = 3600,
        influx_client_kw: Optional[Dict[str, Any]] = None,
        longterm_obs_file: Optional[str] = None,
        preprocess_sourcedb_path: Optional[str] = None,
        load_source_footprints: bool = True,
        cal_targets: Optional[List[str]] = None,
        show_hk_pb: bool = False,
    ) -> None:
        self.ctx_path: str = ctx_path
        self.platform: Literal["satp1", "satp2", "satp3", "lat"] = platform
        self.site_url: str = site_url
        self.buffer_time: float = buffer_time
        self.longterm_obs_file: Optional[str] = longterm_obs_file
        self.preprocess_sourcedb_path: Optional[str] = preprocess_sourcedb_path
        self.load_source_footprints: bool = load_source_footprints
        self.show_hk_pb: bool = show_hk_pb

        if cal_targets is None:
            self.cal_targets = ["jupiter", "saturn", "tau_A", "tauA", "cenA", "mars"]
        else:
            self.cal_targets = cal_targets

        if isinstance(start_time, float):
            self.start_time: dt.datetime = dt.datetime.fromtimestamp(start_time)
        elif isinstance(start_time, str):
            self.start_time = dt.datetime.fromisoformat(start_time)
        else:
            self.start_time = start_time

        if isinstance(stop_time, float):
            self.stop_time: dt.datetime = dt.datetime.fromtimestamp(stop_time)
        elif isinstance(stop_time, str):
            self.stop_time = dt.datetime.fromisoformat(stop_time)
        else:
            self.stop_time = stop_time

        self.hk_cfg = hk_cfg

        if influx_client_kw is None:
            self.influx_client_kw: Dict[str, Any] = {}
        else:
            self.influx_client_kw = influx_client_kw

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "ReportDataConfig":
        return cls(**data)

    @classmethod
    def from_yaml(cls, path: str) -> "ReportDataConfig":
        with open(path, "r") as f:
            return cls.from_dict(yaml.safe_load(f))


@dataclass
class ObsInfo:
    obs_id: str
    start_time: float
    stop_time: float
    wafer_slots_list: str
    stream_ids_list: str
    obs_type: str
    obs_subtype: str
    obs_tube_slot: str
    obs_tags: str = ""
    pwv: float = np.nan
    num_valid_dets: str = ""

    @classmethod
    def from_obsdb_entry(cls, data) -> "ObsInfo":
        obs_info = cls(
            obs_id=data["obs_id"],
            start_time=data["start_time"],
            stop_time=data["stop_time"],
            wafer_slots_list=data["wafer_slots_list"],
            stream_ids_list=data["stream_ids_list"],
            obs_type=data["type"],
            obs_subtype=data["subtype"],
            obs_tube_slot=data["tube_slot"],
        )
        return obs_info

    def __repr__(self):
        inner = ", ".join(f"{k}={repr(v)}" for k, v in self.__dict__.items())
        return f"ObsInfo({inner})"


def arr_to_obs_list(obs_arr: np.ndarray) -> List[ObsInfo]:
    """
    Convert a npy structured array back to a list of ObsInfo objects.
    """
    obs_list: List[ObsInfo] = []
    for entry in obs_arr:
        kw = {}
        for k, v in obs_arr.dtype.fields.items():
            if v[0].type == np.bytes_:  # Convert back to string
                kw[k] = entry[k].decode()
            else:  # Is a numeric type
                kw[k] = entry[k]
        obs_list.append(ObsInfo(**kw))
    return obs_list


def obs_list_to_arr(obs_list: List[ObsInfo]) -> np.ndarray:
    """
    Convert a list of ObsInfo objects to a numpy structured array for processing
    and storage.
    """
    dtype = []
    data = []
    for field in fields(ObsInfo):
        if field.type is str:
            typ: Any = "|S100"
        else:
            typ = field.type
        dtype.append((field.name, typ))
    for obs in obs_list:
        data.append(tuple(getattr(obs, name) for name, _ in dtype))
    return np.array(data, dtype=dtype)


def get_apex_data(cfg: ReportDataConfig):
    """
    Load APEX pwv data within the time range in the ReportDataConfig.
    """
    APEX_DATA_URL = 'http://archive.eso.org/wdb/wdb/eso/meteo_apex/query'

    request = requests.post(APEX_DATA_URL, data={
            'wdbo': 'csv/download',
            'max_rows_returned': 79400,
            'start_date': cfg.start_time.strftime('%Y-%m-%dT%H:%M:%S') + '..' \
                + cfg.stop_time.strftime('%Y-%m-%dT%H:%M:%S'),
            'tab_pwv': 'on',
            'shutter': 'SHUTTER_OPEN',
            #'tab_shutter': 'on',
        })

    # check if any data was found
    lines = request.text.splitlines()
    found_data = False
    for line in lines:
        l = line.strip()
        if l and not l.startswith("#"):
            found_data = True
            break

    if not found_data:
        return None

    def date_converter(d):
        naive_dt = dt.datetime.fromisoformat(d)
        return naive_dt.replace(tzinfo=dt.timezone.utc)

    data = np.genfromtxt(
        StringIO(request.text),
        delimiter=',', skip_header=2,
        converters={0: date_converter},
        dtype=[('dates', dt.datetime), ('pwv', float)],
    )

    outdata = {'timestamps':[d.timestamp() for d in data['dates']],
               'pwv':data['pwv']}
    return outdata


def load_pwv(cfg: ReportDataConfig) -> hkdb.HkResult:
    """
    Load PWV data from the range specified in the ReportDataConfig.
    Uses hk_cfg file or dict specified in the config.
    """

    # don't try and load pwv earlier than 90 days ago
    if cfg.start_time < dt.datetime.now(dt.timezone.utc) - dt.timedelta(days=90):
        return None

    if isinstance(cfg.hk_cfg, str):
        hk_cfg: HkConfig = HkConfig.from_yaml(cfg.hk_cfg)
    elif isinstance(cfg.hk_cfg, dict):
        hk_cfg = HkConfig.from_dict(cfg.hk_cfg)
    else:
        hk_cfg = cfg.hk_cfg

    result = hkdb.load_hk(
        hkdb.LoadSpec(
            cfg=hk_cfg,
            start=cfg.start_time.timestamp() - cfg.buffer_time,
            end=cfg.stop_time.timestamp() + cfg.buffer_time,
            fields=["pwv"],
        ),
        show_pb=cfg.show_hk_pb,
    )

    return result


def get_hk_and_pwv_data(cfg: ReportDataConfig):
    """
    Load the pwv from either the CLASS or APEX radiometer.
    Merge both datasets when available.
    """
    result = load_pwv(cfg)
    result_apex = get_apex_data(cfg)
    if result_apex is not None:
        result_apex = (np.array(result_apex['timestamps']), np.array(0.03+0.84 * result_apex['pwv']))

    if hasattr(result, 'pwv'):
        if result_apex is not None:
            combined_times = np.concatenate((result.pwv[0], result_apex[0]))
            combined_data = np.concatenate((result.pwv[1], result_apex[1]))

            sorted_indices = np.argsort(combined_times)
            sorted_times = combined_times[sorted_indices]
            sorted_data = combined_data[sorted_indices]
            return (sorted_times, sorted_data)
        else:
            return result
    elif result_apex is not None:
        return result_apex
    else:
        return None


def load_qds_data(cfg: ReportDataConfig) -> pd.DataFrame:
    """
    Loads QDS data from influxdb.

    """
    client = InfluxDBClient(**cfg.influx_client_kw)
    buff_time = dt.timedelta(seconds=cfg.buffer_time)

    t0_str = (cfg.start_time - buff_time).isoformat().replace("+00:00", "Z")
    t1_str = (cfg.stop_time + buff_time).isoformat().replace("+00:00", "Z")

    keys = ['time', 'num_valid_dets', 'wafer.bandpass']#, '"wafer_slot"::tag', '"wafer.bandpass"::tag']

    query = f"""
        SELECT """ + ", ".join(keys) +  f""" from "autogen"."preprocesstod" WHERE (
            "tel_tube"::tag = '{cfg.platform}'
            AND time >= '{t0_str}'
            AND time <= '{t1_str}'
        )
    """

    result = client.query(query)
    df = pd.DataFrame(result.get_points())
    missing_keys = [key for key in keys if key not in df]

    if missing_keys:
        logger.warn(f"missing keys from qds: {missing_keys}")
        return None

    df["time"] = pd.to_datetime(df["time"])
    df["timestamp"] = df["time"].apply(lambda x: x.timestamp())

    return df


def merge_qds_and_obs_list(df: pd.DataFrame, obs_list: List[ObsInfo]) -> None:
    timestamps = np.array([o.start_time for o in obs_list])
    obsids = [o.obs_id for o in obs_list]

    def find_obsid(ts: float) -> Optional[str]:
        idx = np.argmin(np.abs(ts - timestamps))
        if np.isclose(ts, timestamps[idx], atol=0.1):
            return obsids[idx]
        else:
            return ""

    df["obs_id"] = df["timestamp"].apply(find_obsid)

    totals = (
        df[["num_valid_dets", "obs_id", "wafer.bandpass"]]
        .groupby(["obs_id", "wafer.bandpass"])["num_valid_dets"]
        .sum()
        .reset_index()
    )

    totals_dict = (
        totals.groupby("obs_id")
              .apply(lambda g: dict(zip(g["wafer.bandpass"], g["num_valid_dets"])))
              .to_dict()
    )

    for obs_id, band_totals in totals_dict.items():
        band_totals.pop("NC", None)
        if not obs_id or obs_id not in obsids:
            continue
        obs_entry = obs_list[obsids.index(obs_id)]
        obs_entry.num_valid_dets = str(band_totals)


@dataclass
class Footprint:
    """
    Class to contain calibration footprint information.

    Attributes
    -------------
    obs_id: str
        Observation ID of the scan.
    target: str
        Calibration target of the footprint.
    xi_p: np.ndarray
        Xi of the source relative to the boresight over the course of the scan.
    eta_p: np.ndarray
        Eta of the source relative to the boresight over the course of the scan.
    bounds: np.ndarray
        Array of coordinates that define the boundary of the footprint.
        This is what will be plotted, as the full source coordinates is too much
        data. This is computed by taking the alpha-shape of the source
        coordinates.
    """

    wafer: str
    source: str
    count: int
    obsids: List[str]

    def to_h5(self, group: h5py.Group):
        """
        Store a footprint in an hdf5 group.
        """
        group.attrs["target"] = self.target
        group.attrs["obs_id"] = self.obs_id
        group.create_dataset("wafers", data=','.join(wafers.astype(str)))

    @classmethod
    def from_h5(cls, group: h5py.Group) -> "Footprint":
        """Convert from an hdf5 group into a footprint object"""
        fp = cls(
            obs_id=group.attrs["obs_id"],
            target=group.attrs["target"],
            wafers=np.array(group["wafers"]),
        )
        return fp


@dataclass
class ReportData:
    """
    High-level summary data of the region of time specified in the ReportConfig.

    Attributes
    ------------
    cfg: ReportConfig
        Configuration object used to generate the report
    obs_list: List[ObsInfo]
        List of observation info for each observation in the configured time range.
        This contains data from the ObsDb, HkDb, and QDS database.
    pwv: np.ndarray
        PWV throughout the specified time range. This is a 2D array where
        the first element is an array of timestamps, and the second element is
        an array of PWV values. This is pulled from the HKDB.
    cal_footprint: List[Footprint]
        List of Footprint objects which describe source footprint found during

    """

    cfg: ReportDataConfig
    obs_list: List[ObsInfo]
    pwv: np.ndarray
    source_footprints: Optional[List[Footprint]] = None
    longterm_obs_df: Optional[pd.DataFrame] = None

    @classmethod
    def build(cls, cfg: ReportDataConfig) -> "ReportData":
        ctx = Context(cfg.ctx_path)

        obs_list = [
            ObsInfo.from_obsdb_entry(o)
            for o in ctx.obsdb.query(
                f"start_time >= {cfg.start_time.timestamp()} and "
                f"start_time <= {cfg.stop_time.timestamp()}"
            )
        ]

        #for i, o in tqdm(enumerate(obs_list), total=len(obs_list)):
        for i, o in enumerate(obs_list):
            o.obs_tags = ",".join(ctx.obsdb.get(o.obs_id, tags=True)['tags'])

        if cfg.longterm_obs_file is not None:
            longterm_obs_df = cls.load(cfg.longterm_obs_file)
        else:
            longterm_obs_df = None

        logger.info("Loading PWV data")
        pwv = get_hk_and_pwv_data(cfg)

        logger.info("Loading QDS data")
        qds_df = load_qds_data(cfg)

        if qds_df is not None:
            logger.info("Merging PWV and QDS data with obs list")
            merge_qds_and_obs_list(qds_df, obs_list)
        else:
            logger.warn("QDS data not found")

        # Add PWV data to obs_list
        if pwv is not None:
            for o in obs_list:
                m = np.logical_and.reduce([pwv[0] >= o.start_time, pwv[0] <= o.stop_time])
                _pwv = np.nanmean(pwv[1][m])
                if -0.1 < _pwv < 3.5:
                    o.pwv = _pwv
        else:
            logger.warn("pwv data not found")
            for o in obs_list:
                o.pwv = -9999.

        data: "ReportData" = cls(
            cfg=cfg, obs_list=obs_list, pwv=pwv, longterm_obs_df=longterm_obs_df
        )

        if cfg.load_source_footprints:
            logger.info("Loading Source Footprints")
            source_footprints = get_source_footprints(data)
            data.source_footprints = source_footprints
        return data

    def save(self, path: str) -> None:
        """
        Save compiled data to an H5 file.
        """
        with h5py.File(path, "w") as hdf:
            d = self.cfg.__dict__
            for k, v in d.items():
                if isinstance(v, dt.datetime):
                    d[k] = v.isoformat()
                else:
                    try:
                        d[k] = v
                    except:
                        raise Exception(f"Key: {k} cannot be converted to h5.")
            hdf.attrs["cfg"] = json.dumps(d)
            hdf.create_dataset("pwv", data=self.pwv)
            hdf.create_dataset("obs_list", data=obs_list_to_arr(self.obs_list))

            if self.source_footprints is not None:
                fp_grp = hdf.create_group("source_footprints")
                wafers = np.array([fp.wafer for fp in self.source_footprints], dtype='S')
                sources = np.array([fp.source for fp in self.source_footprints], dtype='S')
                counts = np.array([fp.count for fp in self.source_footprints], dtype='i4')
                obsids = np.array([','.join(fp.obsids).encode('utf-8') for fp in self.source_footprints], dtype=h5py.special_dtype(vlen=bytes))
                fp_grp.create_dataset('wafer', data=wafers)
                fp_grp.create_dataset('source', data=sources)
                fp_grp.create_dataset('count', data=counts)
                fp_grp.create_dataset('obsids', data=obsids)
                # for i, fp in enumerate(self.source_footprints):
                #     grp = fp_grp.create_group(f"fp_{i}")
                #     grp.attrs["target"] = fp.target
                #     grp.attrs["obs_id"] = fp.obs_id
                #     grp.create_dataset("wafers", data=','.join(wafers.astype(str)))

    @classmethod
    def load(cls, path: str) -> "ReportData":
        with h5py.File(path, "r") as hdf:
            # Load config
            cfg = ReportDataConfig(**json.loads(hdf.attrs["cfg"]))

            obs_list = arr_to_obs_list(hdf["obs_list"])
            pwv = np.array(hdf["pwv"])

            if "source_footprints" in hdf:
                fps = []
                fp_grp = hdf["source_footprints"]
                wafers = [w.decode() for w in fp_grp['wafer'][()]]
                sources = [s.decode() for s in fp_grp['source'][()]]
                counts = fp_grp['count'][()]
                obsids_list = [o.decode().split(',') for o in fp_grp['obsids'][()]]

                for wafer, source, count, obsids in zip(wafers, sources, counts, obsids_list):
                    fps.append(Footprint(
                        wafer=wafer, source=source, count=count, obsids=obsids
                    ))
                # fps = [Footprint.from_h5(fp) for fp in hdf["source_footprints"].values()]
            else:
                fps = None

        return cls(
            cfg=cfg,
            obs_list=obs_list,
            pwv=pwv,
            source_footprints=fps,
        )


def get_source_footprints(d: ReportData) -> List[Footprint]:
    fps: List[Footprint] = []
    db = ManifestDb(d.cfg.preprocess_sourcedb_path)
    entries = db.inspect()
    source_obs_list = []
    for entry in db.inspect():
        if (float(entry['obs:obs_id'].split('_')[1]) >= d.cfg.start_time.timestamp()) & \
        (float(entry['obs:obs_id'].split('_')[1]) <= d.cfg.stop_time.timestamp()):
            source_obs_list.append(entry['obs:obs_id'])
            
    coverage_data = defaultdict(lambda: defaultdict(lambda: {'count': 0, 'obsids': set()}))
    
    for obs in source_obs_list:
        entry = db.inspect({'obs:obs_id': obs})
        if entry[0]['coverage'] == '':
            continue
        for pair in entry[0]['coverage'].split(','):
            source, wafer = pair.split(':')
            source = source.casefold()
            if source in d.cfg.cal_targets:
                coverage_data[wafer][source]['count'] += 1
                coverage_data[wafer][source]['obsids'].add(obs)
                
    for wafer, sources in coverage_data.items():
        for source, info in sources.items():
            fps.append(Footprint(wafer=wafer,
                                 source=source,
                                 count=info['count'],
                                 obsids=sorted(info['obsids'])))

    return fps



import numpy as np


def get_apodize_window_for_ends(aman, apodize_samps=1600):
    """
    Generate an apodization window using a cosine taper at the beginning and end.

    Args:
        aman: An axismanager
        apodize_samps (int): Number of samples to apply the cosine taper to at each end.

    Returns:
        numpy.ndarray: An array representing the apodization window.
    """
    w = np.ones(aman.samps.count)
    cosedge = np.cos(np.linspace(0, np.pi/2, apodize_samps))
    w[-apodize_samps:] = cosedge
    w[:apodize_samps] = np.flip(cosedge)
    return w


def get_apodize_window_from_flags(aman, flags, apodize_samps=200):
    """
    Generate an apodization window based on flag values. Apply cosine tapering every 
    continuous portion of data between flagged region.

    Args:
        aman: An axismanager
        flags (str or RangesMatrix or Ranges): Flags of mask in RangesMatrix/Ranges. If provided by 
            a string, 'aman.flags[flags]' is used for the flags.
        apodize_samps (int): Number of samples to apply the cosine taper.

    Returns:
        numpy.ndarray: An array representing the apodization window.
    """
    if isinstance(flags, str):
        flags = aman.flags[flags]
    flags_mask = flags.mask()

    if flags_mask.ndim == 1:
        flag_is_1d = True
    else:
        all_columns_same = np.all(np.all(flags_mask == flags_mask[0, :], axis=0))
        if all_columns_same:
            flag_is_1d = True
            flags_mask = flags_mask[0]
        else:
            flag_is_1d = False

    apodizer = ~flags_mask
    apodizer = apodizer.astype(float)
    cosedge = np.cos(np.linspace(0, np.pi/2, apodize_samps))

    if flag_is_1d:
        idxes_left = np.where(np.diff(apodizer) == -1)[0]
        idxes_right = np.where(np.diff(apodizer) == 1)[0]

        for _left in idxes_left:
            _apo_idxes_left = (_left-apodize_samps+1, _left+1)
            if _apo_idxes_left[0] < 0:
                apodizer[:_apo_idxes_left[1]] = 0
            else:
                apodizer[_apo_idxes_left[0]:_apo_idxes_left[1]] *= cosedge

        for _right in idxes_right:
            _apo_idxes_right = (_right-1, _right+apodize_samps-1)
            if _apo_idxes_right[1] > aman.samps.count - 1:
                apodizer[_apo_idxes_right[0]:] = 0
            else:
                apodizer[_apo_idxes_right[0]:_apo_idxes_right[1]] *= np.flip(cosedge)
    else:
        for di in range(aman.dets.count):
            idxes_left = np.where(np.diff(apodizer[di]) == -1)[0]
            idxes_right = np.where(np.diff(apodizer[di]) == 1)[0]

            for _left in idxes_left:
                _apo_idxes_left = (_left-apodize_samps+1, _left+1)
                if _apo_idxes_left[0] < 0:
                    apodizer[di][:_apo_idxes_left[1]] = 0
                else:
                    apodizer[di][_apo_idxes_left[0]:_apo_idxes_left[1]] *= cosedge

            for _right in idxes_right:
                _apo_idxes_right = (_right-1, _right+apodize_samps-1)
                if _apo_idxes_right[1] > aman.samps.count - 1:
                    apodizer[di][_apo_idxes_right[0]:] = 0
                else:
                    apodizer[di][_apo_idxes_right[0]:_apo_idxes_right[1]] *= np.flip(cosedge)

    return apodizer


def apodize_cosine(aman, signal_name='signal', apodize_samps=1600, in_place=True,
                   apo_axis='apodized', window=None):
    """
    Function to smoothly filter the timestream to 0's on the ends with a
    cosine function. If window is provided, multiply the window function to 
    aman[signal_name].

    Args:
        signal_name (str): Axis to apodize
        apodize_samps (int): Number of samples on tod ends to apodize.
        in_place (bool): writes over signal with apodized version
        apo_axis (str): Axis to store the apodized signal if not in place.
        window (numpy.ndarray): Precomputed apodization window.
    """
    if window is None:
        w = get_apodize_window_for_ends(aman, apodize_samps)

    if in_place:
        aman[signal_name] *= w
    else:
        aman.wrap_new(apo_axis, dtype='float32', shape=('dets', 'samps'))
        aman[apo_axis] = aman[signal_name]*w
    return



"""Command-line interface for inspecting Context, ObsFileDb, ObsDb,
ManifestDb.

"""

import argparse
import os
import sys

from .. import context
from . import manifest, obsdb, obsfiledb, ObsDb


CLI_NAME = 'so-metadata'
HELP = """Read context / metadata databases and print summary or detailed
information.  You have to pass a db type (obsdb, obsfiledb, ...)
followed by the a db filename (or the path to a context.yaml, from
which the relevant db will be loaded).

"""

USAGE = "%(prog)s [db_type] [filename]"


def get_parser():
    parser = argparse.ArgumentParser(description=HELP)

    sps = parser.add_subparsers(dest='_subcmd')

    # context
    sp = sps.add_parser('context', description=
                        'Inspect a context file and print out a few things about it.')
    sp.add_argument('ctx_file', help='Path to Context yaml file.')

    # obsdb
    sp = sps.add_parser('obsdb', description=
                        'Inspect an ObsDb.  By default, prints out summary information.  '
                        'Pass --list to see a list of all observations; or pass an obs_id '
                        'to get detailed info about that obs.')
    sp.add_argument('db_file', help='Path to database (or context.yaml).')
    sp.add_argument('--type', choices=['context', 'db'], help=
                    'Specifies how to interpret the file (to override guessing '
                    'based on the filename extension.')
    sp.add_argument('--list', '-l', action='store_true', help=
                    'List all observations')
    sp.add_argument('--query', '-q', default='1', help=
                    'Restrict listed items with an ObsDb query string.')
    sp.add_argument('--key', '-k', action='append', default=[], help=
                    'If listing observations, also include the speficied db fields '
                    'in addition to obs_id.')
    sp.add_argument('--tag', '-t', action='append', default=[], help=
                    'Add a tag as an eligible column for query.')
    sp.add_argument('obs_id', nargs='?')

    # obsfiledb
    sp = sps.add_parser('obsdb-diff', description=
                        'Compare obsdb to an upstream copy and maybe patch it to match.')
    sp.add_argument('db_file', help='Path to sqlite3 obsdb.')
    sp.add_argument('upstream_db')
    sp.add_argument('--patch', action='store_true')

    # obsfiledb
    sp = sps.add_parser('obsfiledb', description=
                        'Inspect an ObsFileDb.  This can be used to list all files, '
                        'and to perform batch updates of filenames.')
    obsfiledb.get_parser(sp)

    # metadata
    sp = sps.add_parser('metadata', description="Inspect (or modify) a ManifestDb.")
    manifest.get_parser(sp)

    return parser


def _set_type(args, parser):
    if args.type is None:
        ext = os.path.splitext(args.db_file)[1]
        if ext in ['.yaml', '.yml']:
            args.type = 'context'
        elif ext in ['.sqlite', '.db']:
            args.type = 'db'
        else:
            parser.error(
                f'Not sure how to decode file with extension "{ext}"; pass --type=...')
    return args.type


def _all_tokens(args, subdelim=','):
    # ['a', 'b,c', 'd'] -> ['a', 'b', 'c', 'd']
    out = []
    for a in args:
        out.extend([_x.strip() for _x in a.split(subdelim)])
    return out


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    parser = get_parser()
    args = parser.parse_args(args=args)

    if args._subcmd == 'context':
        ctx = context.Context(args.ctx_file)
        base_path = os.path.split(ctx.filename)[0]

        print('Resolving DB paths:')
        for db in ['obsfiledb', 'obsdb', 'detdb']:
            if ctx.get(db) is not None:
                db_path = os.path.abspath(os.path.join(base_path, ctx[db]))
            else:
                db_path = '(none declared)'
            print(f'  {db:<10}: {db_path}')
        print()
        print('Resolving metadata paths:')
        for entry in ctx.get('metadata', []):
            if entry.get('db'):
                db_path = os.path.abspath(os.path.join(base_path, entry['db']))
            else:
                db_path = '(no db filename)'
            name = entry.get('name')
            if name is None:
                name = '(unnamed)'
            if not isinstance(name, str):  # it could be a list!
                name = str(name)
            if entry.get('det_info'):
                name += ' -- [det_info]'
            print(f'  {name:<30}: {db_path}')
        print()

    elif args._subcmd == 'obsdb':
        _set_type(args, parser)

        if args.type == 'context':
            ctx = context.Context(args.db_file)
            db = ctx.obsdb
        elif args.type == 'db':
            db = obsdb.ObsDb(args.db_file)

        # Summary mode?
        if args.obs_id is None:
            if args.list:
                tags = _all_tokens(args.tag)

                rows = db.query(args.query, tags=tags)

                if args.list:
                    fields = ['obs_id'] + _all_tokens(args.key)
                    for line in rows:
                        print(' '.join([str(line[k]) for k in fields]))
            else:
                info = db.info()
                print('  ObsDb summary')
                print('=================')
                print()
                if args.query != '1':
                    print('Note the "--query" argument is not applied to these results!')
                    print()
                print('There are %i observations' % info['count'])
                print()
                print(' Index fields (distinct values, examples)')
                print('------------------------------------------')
                for f, (count, eg_str) in info['fields'].items():
                    print(f'  {f:<15} ({count:d} distinct), e.g.: {eg_str}')
                print()
                print(' Tags defined (incidence)')
                print('--------------------------')
                for t, count in info['tags'].items():
                    print(f'  {t:<15} ({count:d} assignments)')
                print()
        else:
            print(' Observation info')
            print('-' * 60)
            item = db.get(args.obs_id)
            if item is None:
                print(f'  "{args.obs_id}" not found!')
            else:
                for k, v in db.get(args.obs_id, tags=True).items():
                    print(f'  {k:<20}: {v}')

    elif args._subcmd == 'obsdb-diff':
        print(f'Comparing to {args.upstream_db} ...')
        db = ObsDb(args.db_file)
        db_right = ObsDb(args.upstream_db)
        report = obsdb.diff_obsdbs(db, db_right)
        if not report['different']:
            print(' ... databases are in sync.')
        elif report['patchable']:
            print(' ... upstream is different, but the target db can be patched to match.')
        else:
            print(' ... upstream and target have irreconcilable differences.')
            parser.exit(1)

        if args.patch and report['different']:
            print()
            print('Patching ...')
            obsdb.patch_obsdb(report['patch_data'], db)
            print(' ... done')
            print()

    elif args._subcmd == 'obsfiledb':
        obsfiledb.main(args, parser=parser)

    elif args._subcmd == 'metadata':
        manifest.main(args)

    else:
        parser.error(f'Provide a subcmd.')

if __name__ == '__main__':
    main()




# Copyright (c) 2023-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Template regression mapmaking.
"""
import os

import numpy as np
from astropy import units as u
import toast
from toast.mpi import flatten
import toast.ops
from toast.observation import default_values as defaults

from .. import ops as so_ops
from .job import workflow_timer
from .proc_noise_est import select_mapmaking_noise_model


def setup_splits(operators):
    """Add commandline args and operators for SAT mapmaking splits.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(so_ops.Splits(name="splits", enabled=False))


@workflow_timer
def splits(job, otherargs, runargs, data):
    """Apply mapmaking splits.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    job_ops = job.operators
    splits = job.operators.splits

    if splits.enabled:
        mapmaker_select_noise_and_binner(job, otherargs, runargs, data)
        if job_ops.mapmaker.enabled:
            splits.mapmaker = job_ops.mapmaker
        elif job_ops.filterbin.enabled:
            splits.mapmaker = job_ops.filterbin
        else:
            msg = "No mapmaker is enabled!"
            raise RuntimeError(msg)
        splits.output_dir = splits.mapmaker.output_dir
        mapmaker_run(job, otherargs, runargs, data, splits)


def setup_mapmaker(operators, templates):
    """Add commandline args, operators, and templates for TOAST mapmaker.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    templates.append(
        toast.templates.Offset(
            name="baselines",
            step_time=2.0 * u.second,
            enabled=False,
        )
    )
    templates.append(
        toast.templates.Periodic(
            name="azss",
            key=defaults.azimuth,
            flags=defaults.shared_flags,
            flag_mask=defaults.shared_mask_invalid,
            increment=np.pi / 180.0,  # One degree, az field is in radians
            bins=None,
            enabled=False,
        )
    )
    templates.append(
        toast.templates.Fourier2D(
            name="fourier2d",
            correlation_length=5.0 * u.second,
            correlation_amplitude=10.0,
            order=1,
            fit_subharmonics=False,
        )
    )
    operators.append(toast.ops.BinMap(name="binner", pixel_dist="pix_dist"))
    operators.append(
        toast.ops.BinMap(
            name="binner_final", enabled=False, pixel_dist="pix_dist_final"
        )
    )
    operators.append(toast.ops.MapMaker(name="mapmaker", det_data=defaults.det_data))


def mapmaker_select_noise_and_binner(job, otherargs, runargs, data):
    """Helper function to setup noise model and binner for mapmapking.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    log = toast.utils.Logger.get()

    # Configured templates for this job
    job_tmpls = job.templates

    # Configured operators for this job
    job_ops = job.operators

    if job_ops.mapmaker.enabled:
        job_ops.mapmaker.binning = job_ops.binner
        job_ops.mapmaker.template_matrix = toast.ops.TemplateMatrix(
            templates=[job_tmpls.baselines, job_tmpls.azss]
        )
        job_ops.mapmaker.map_binning = job_ops.binner_final
        job_ops.mapmaker.output_dir = otherargs.out_dir
        tmsg = "  "
        for tmpl in job_ops.mapmaker.template_matrix.templates:
            estr = "(enabled)" if tmpl.enabled else "(disabled)"
            tmsg += f"{tmpl.name} {estr}, "
        log.info_rank(
            "  Using regression templates:",
            comm=data.comm.comm_world,
        )
        log.info_rank(
            tmsg,
            comm=data.comm.comm_world,
        )
        # Noise model.  If noise estimation is not enabled, and no existing noise model
        # is found, then create a fake noise model with uniform weighting.
        noise_model = select_mapmaking_noise_model(job, otherargs, runargs, data)
        if noise_model is None:
            for ob in data.obs:
                (estrate, _, _, _, _) = toast.utils.rate_from_times(
                    ob.shared[defaults.times].data
                )
                ob["fake_noise"] = toast.noise_sim.AnalyticNoise(
                    detectors=ob.all_detectors,
                    rate={x: estrate * u.Hz for x in ob.all_detectors},
                    fmin={x: 1.0e-5 * u.Hz for x in ob.all_detectors},
                    fknee={x: 0.0 * u.Hz for x in ob.all_detectors},
                    alpha={x: 1.0 for x in ob.all_detectors},
                    NET={
                        x: 1.0 * u.K * np.sqrt(1.0 * u.second)
                        for x in ob.all_detectors
                    },
                )
            log.info_rank(
                "  Using fake noise model with uniform weighting",
                comm=data.comm.comm_world,
            )
            noise_model = "fake_noise"
        job_ops.binner.noise_model = noise_model
        job_ops.binner_final.noise_model = noise_model

        if job_tmpls.baselines.enabled:
            job_tmpls.noise_model = noise_model


@workflow_timer
def mapmaker_run(job, otherargs, runargs, data, map_op):
    """Run a mapmaker, optionally per observation.

    This runs the mapmaker either in single shot or per
    detector/observation.  Currently this supports instances of the
    `Mapmaker` and `Splits` operators.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.
        map_op (Operator):  The operator to run.

    Returns:
        None

    """
    log = toast.utils.Logger.get()

    if map_op.enabled:
        do_obsmaps = hasattr(otherargs, "obsmaps") and otherargs.obsmaps
        do_detmaps = hasattr(otherargs, "detmaps") and otherargs.detmaps
        do_intervalmaps = (
            hasattr(otherargs, "intervalmaps") and otherargs.intervalmaps
        )
        # See if user wants separate detector maps
        if do_detmaps:
            my_dets = data.all_local_detectors(flagmask=defaults.det_mask_invalid)
            if data.comm.comm_world is None:
                all_dets = my_dets
            else:
                all_dets = data.comm.comm_world.allgather(my_dets)
                all_dets = sorted(set(flatten(all_dets)))
        else:
            all_dets = [None]
        if do_obsmaps and do_intervalmaps:
            log.warning_rank(
                "--intervalmaps overrides --obsmaps", data.comm.comm_world
            )
        if do_obsmaps or do_intervalmaps:
            log.debug_rank(
                f"{data.comm.group}: Running observation or interval maps",
                comm=data.comm.comm_world,
            )

        mapmaker_name = map_op.name
        for det in all_dets:
            if det is None:
                # Map all detectors together
                detectors = None
            else:
                # Single detector mode, append detector name to all
                # data products
                map_op.name = f"{mapmaker_name}_{det}"
                detectors = [det]
            if do_obsmaps or do_intervalmaps:
                # Map each observation separately
                timer_obs = toast.timing.Timer()
                timer_obs.start()
                group = data.comm.group
                orig_name = map_op.name
                orig_outdir = map_op.output_dir
                orig_comm = data.comm
                new_comm = toast.Comm(world=data.comm.comm_group)
                for iobs, obs in enumerate(data.obs):
                    log.info_rank(
                        f"{group} : mapping observation {iobs + 1} "
                        f"/ {len(data.obs)}.",
                        comm=new_comm.comm_world,
                    )
                    # Set the observation name used in output file names.
                    # This is the base observation name without any "demod_"
                    # string.
                    obs_str = obs.name.replace("demod_", "")
                    # Data object that only covers one observation
                    obs_data = data.select(obs_uid=obs.uid)
                    # Set the output directory to the session name
                    map_op.output_dir = os.path.join(orig_outdir, obs.session.name)

                    # Delete any solver or other products
                    import re
                    for k in list(obs_data.keys()):
                        if re.match(r".*_solve_.*", k) is not None:
                            del obs_data[k]

                    # Replace comm_world with the group communicator
                    obs_data._comm = new_comm
                    if isinstance(map_op, so_ops.Splits):
                        binner = map_op.mapmaker.binning
                    else:
                        binner = map_op.binning

                    if binner.pixel_pointing.pixels in obs:
                        del obs[binner.pixel_pointing.pixels]
                    if binner.stokes_weights.weights in obs:
                        del obs[binner.stokes_weights.weights]
                    if binner.pixel_pointing.detector_pointing.quats in obs:
                        del obs[binner.pixel_pointing.detector_pointing.quats]

                    orig_view = binner.pixel_pointing.view
                    if do_intervalmaps and orig_view is not None:
                        if isinstance(map_op, so_ops.Splits):
                            msg = "Interval mapping cannot be used with Splits"
                            raise RuntimeError(msg)
                        # Map each interval separately
                        ob = obs_data.obs[0]
                        times = ob.shared[defaults.times].data
                        views = ob.intervals[orig_view]
                        for iview, view in enumerate(views):
                            # Add a view for this specific interval
                            single_view = f"{orig_view}-{iview}"
                            ob.intervals[single_view] = toast.IntervalList(
                                times, timespans=[(view.start, view.stop)]
                            )
                            binner.pixel_pointing.view = single_view
                            map_op.name = f"{orig_name}_{obs_str}-{iview}"
                            map_op.reset_pix_dist = True
                            try:
                                map_op.apply(obs_data, detectors=detectors)
                                log.info_rank(
                                    f"{group} : Mapped det={det} "
                                    f"{obs_str}-{iview} / {len(views)} in",
                                    comm=new_comm.comm_world,
                                    timer=timer_obs,
                                )
                            except Exception as e:
                                log.info_rank(
                                    f"{group} : Failed to map "
                                    f"{obs_str}-{iview} / {len(views)} (e) in",
                                    comm=new_comm.comm_world,
                                    timer=timer_obs,
                                )
                        binner.pixel_pointing.view = orig_view
                    else:
                        # Map the observation as a whole
                        # Rename the operator with the observation suffix
                        map_op.name = f"{orig_name}_{obs_str}"
                        if isinstance(map_op, so_ops.Splits):
                            # Reset the pixel distribution of the underlying
                            # mapmaker
                            map_op.mapmaker.reset_pix_dist = True
                        else:
                            # Reset the trait on this mapmaker
                            map_op.reset_pix_dist = True

                    # Map this observation
                    map_op.apply(obs_data, detectors=detectors)

                    log.info_rank(
                        f"{group} : Mapped det={det} obs={obs_str} in",
                        comm=new_comm.comm_world,
                        timer=timer_obs,
                    )
                log.info_rank(
                    f"{group} : Done mapping {len(data.obs)} observations.",
                    comm=new_comm.comm_world,
                )
                map_op.name = orig_name
                map_op.output_dir = orig_outdir
                data._comm = orig_comm
                del new_comm
            else:
                log.debug_rank(
                    f"{data.comm.group}: Calling mapmaker.apply() directly, "
                    f"det={det}",
                    comm=data.comm.comm_world,
                )
                map_op.apply(data, detectors=detectors)


@workflow_timer
def mapmaker(job, otherargs, runargs, data):
    """Run the TOAST mapmaker.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    log = toast.utils.Logger.get()

    # Configured templates for this job
    job_tmpls = job.templates

    # Configured operators for this job
    job_ops = job.operators

    mapmaker_select_noise_and_binner(job, otherargs, runargs, data)
    mapmaker_run(job, otherargs, runargs, data, job_ops.mapmaker)



"""
Generically useful utility functions.
"""
from typing import Optional
import numpy as np
from numpy.typing import NDArray
from so3g import block_moment, block_moment64


def get_block_moment(
    tod: NDArray[np.floating],
    block_size: int,
    moment: int = 1,
    central: bool = True,
    shift: int = 0,
    output: Optional[NDArray[np.floating]] = None,
) -> NDArray[np.floating]:
    """
    Compute the n'th moment of data in blocks along each row.
    Note that the blocks are made to be exclusive,
    so any samples left at the end will be in a smaller standalone block.
    This is a wrapper around ``so3g.block_moment``.

    Arguments:

        tod: Data to compute the moment of.
             Should  be (ndet, nsamp) or (nsamp).
             Must be float32 or float64.

        block_size: Size of block to use.

        moment: Which moment to compute.
                Must be >= 1.

        central: If True compute the mean centered moment.

        shift: Sample to start the blocks at, will be 0 before this.

        output: Array to put the blocked moment into.
                If provided must be the same shape as tod.
                If None, will be intialized from tod.
    Returns:

        block_moment: The blocked moment.
                      Will have the same shape as tod.
                      If output is provided it is modified in place and retured here.
    """
    if not np.any(np.isfinite(tod)):
        raise ValueError("Only finite values allowed in tod")
    orig_shape = tod.shape
    dtype = tod.dtype.name
    tod = np.atleast_2d(tod)
    if len(tod.shape) > 2:
        raise ValueError("tod may not have more than 2 dimensions")
    if dtype not in ["float32", "float64"]:
        raise TypeError("tod must be float32 or float64")

    if output is None:
        output = np.ascontiguousarray(np.empty_like(tod))
    if output.shape != tod.shape:
        raise ValueError("output shape does not match tod")
    if output.dtype.name != dtype:
        raise TypeError("output type does not match tod")

    if moment < 1:
        raise ValueError("moment must be at least 1")

    if dtype == "float32":
        block_moment(tod, output, block_size, moment, central, shift)
    else:
        block_moment64(tod, output, block_size, moment, central, shift)

    return output.reshape(orig_shape)



import os
import yaml
import logging
import shutil
import numpy as np
import datetime as dt
from sqlalchemy import or_, and_, not_
from collections import OrderedDict

from .load_smurf import (
    TimeCodes,
    SupRsyncType,
    Finalize,
    SmurfStatus,
    logger as smurf_log
)
from .imprinter import (
    Books,
    Imprinter,
    BOUND,
    UNBOUND,
    UPLOADED,
    FAILED,
    WONT_BIND,
    DONE,
    SMURF_EXCLUDE_PATTERNS,
)
import sotodlib.io.imprinter_utils as utils
from .imprinter_cli import autofix_failed_books
from .datapkg_utils import walk_files, just_suprsync

from .bookbinder import log as book_logger

def combine_loggers(imprint, fname=None):
    log_list = [imprint.logger, smurf_log, book_logger]
    logger = logging.getLogger("DataPackaging")
    logger.setLevel(logging.DEBUG)

    formatter = logging.Formatter(
        '%(levelname)s - %(name)s - %(message)s'
    )
    # Create a file handler
    if fname is not None:
        handler = logging.FileHandler(fname)
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        [l.addHandler(handler) for l in log_list]  

    # Create a stream handler to print logs to the console
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.INFO)  # You can set the desired log level for console output
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)
    return logger

class DataPackaging:
    def __init__(self, platform, log_filename=None):
        self.platform = platform
        self.imprint = Imprinter.for_platform(platform)
        self.logger = combine_loggers(self.imprint, fname=log_filename)
        self.session = self.imprint.get_session()
        if self.imprint.build_det:
            self.g3session, self.SMURF = self.imprint.get_g3tsmurf_session(return_archive=True)
        else: 
            self.g3session = None
            self.SMURF = None
        self.HK = self.imprint.get_g3thk()

    def get_first_timecode_on_disk(self, include_hk=True):
        tc = 50000
        if self.imprint.build_det:
            tc = min([
                tc,
                int(sorted(os.listdir(self.SMURF.meta_path))[0]),
                int(sorted(os.listdir(self.SMURF.archive_path))[0]),
            ])
        if include_hk:
            tc = min([
                tc,
                int(sorted(os.listdir(self.HK.hkarchive_path))[0])
            ])
        if tc == 50000:
            raise ValueError(f"Found no timecode folders for {self.platform}")
        return tc   

    def get_first_timecode_in_staged(self, include_hk=True):
        q = self.session.query(Books).filter(
            Books.status == UPLOADED,
        )
        if not include_hk:
            q = q.filter(Books.type != 'hk')
        first = q.order_by(Books.start).first()
        tc = int( first.start.timestamp() // 1e5)
        return tc   

    def all_files_in_timecode(self, timecode, include_hk=True):
        flist = []
        if self.imprint.build_det:
            stc = os.path.join(self.SMURF.meta_path, str(timecode))
            flist.extend(walk_files(stc, include_suprsync=True))
            ttc = os.path.join(self.SMURF.archive_path, str(timecode))
            flist.extend(walk_files(ttc, include_suprsync=True))
        if include_hk:
            htc = os.path.join(self.HK.hkarchive_path, str(timecode))
            flist.extend(walk_files(htc, include_suprsync=True))
        return flist

    def get_suprsync_files(self, timecode):
        if not self.imprint.build_det:
            return []
        stc = os.path.join(self.SMURF.meta_path, str(timecode))
        ttc = os.path.join(self.SMURF.archive_path, str(timecode))
        flist = []

        if not os.path.exists(stc) and not os.path.exists(ttc):
            return flist
        if os.path.exists(ttc) and 'suprsync' in os.listdir(ttc):
            for root, _, files in os.walk(os.path.join(ttc, 'suprsync')):
                for name in files:
                    flist.append(os.path.join(ttc, root, name))
        if os.path.exists(stc) and 'suprsync' in os.listdir(stc):
            for root, _, files in os.walk(os.path.join(stc, 'suprsync')):
                for name in files:
                    flist.append(os.path.join(stc, root, name))
        return flist

    def check_hk_registered(self, timecode, complete):
        min_ctime = timecode*1e5
        max_ctime = (timecode+1)*1e5

        self.HK.add_hkfiles(
            min_ctime=min_ctime, max_ctime=max_ctime, 
            show_pb=False, update_last_file=False,
        )
        self.imprint.register_hk_books(
            min_ctime=min_ctime, 
            max_ctime=max_ctime, 
        )
        # check the hk book is registered
        book = self.session.query(Books).filter( 
            Books.bid == f"hk_{timecode}_{self.platform}"
        ).one_or_none()
        if book is None:
            complete[0] = False
            complete[1] += f"HK book hk_{timecode}_{self.platform} missing\n"
        elif book.status == UNBOUND:
            try:
                self.imprint.bind_book(book)
            except:
                self.logger.warning(f"Failed to bind {book.bid}")
            if book.status < BOUND:
                complete[0] = False
                complete[1] += f"Book hk_{timecode}_{self.platform} not bound"
        return complete

    def make_timecode_complete(
        self, timecode, try_binding_books=True, try_single_obs=True,
        include_hk=True,
    ):
        """
        Carefully go through an entire timecode and check that the data packaging as
        complete as it can be. The verification will also try and fix any errors
        found in the system. Updating databases, registering books, and binding
        books if try_binding_books is True

        Arguments
        ----------
        timecode: int
            5-digit ctime to check for completion
        try_binding_books: bool
            if true, go through and try to bind any newly registered books
        try_single_obs: bool
            if true, tries to register any missing observations as single wafer
            observations if registering as multi-wafer observations fails. This
            happens sometimes if the stream lengths are very close to the
            minimum overlap time
        include_hk: bool
            if true, also checkes everything related to hk 
        """
        
        complete = [True, ""]
        min_ctime = timecode*1e5
        max_ctime = (timecode+1)*1e5

        if not self.imprint.build_det:
            ## no detector data tracked by imprinter
            if include_hk:
                return self.check_hk_registered(timecode, complete)
            else:
                self.logger.warning(
                    f"No detector data built for platform "
                    f"{self.imprint.daq_node} and not checking HK. Nothing to "
                    "check for completion"
                )
                return complete

        has_smurf, has_timestreams = True, True
        stc = os.path.join(self.SMURF.meta_path, str(timecode))
        ttc = os.path.join(self.SMURF.archive_path, str(timecode))

        if not os.path.exists(stc):
            self.logger.debug(f"TC {timecode}: No level 2 smurf folder")
            has_smurf = False
        if not os.path.exists(ttc):
            self.logger.debug(f"TC {timecode}: No level 2 timestream folder")
            has_timestreams = False

        if os.path.exists(ttc) and just_suprsync(ttc):
            self.logger.info(
                f"TC {timecode}: Level 2 timestreams is only suprsync"
            )
            has_timestreams = False
        if os.path.exists(stc) and just_suprsync(stc):
            self.logger.info(f"TC {timecode}: Level 2 smurf is only suprsync")
            has_smurf = False
        
        if not has_smurf and not has_timestreams:
            return complete
        if not has_smurf and has_timestreams:
            self.logger.error(f"TC {timecode}: Has timestreams folder without smurf!")        
        
        overall_final_ctime = self.SMURF.get_final_time(
            self.imprint.all_slots, check_control=False
        )
        tcode_limit = int(overall_final_ctime//1e5)
        if timecode+1 > tcode_limit:
            raise ValueError(
                f"We cannot check files from {timecode} because finalization time "
                f"is {overall_final_ctime}"
            )

        self.logger.info(f"Checking Timecode {timecode} for completion")
        ## check for files on disk to be in database
        missing_files = self.SMURF.find_missing_files(
            timecode, session=self.g3session
        )
        if len(missing_files) > 0:
            self.logger.warning(
                f"{len(missing_files)} files not in G3tSmurf"
            )
            self.SMURF.index_metadata(
                min_ctime=min_ctime, 
                max_ctime=max_ctime, 
                session=self.g3session
            )
            self.SMURF.index_archive(
                min_ctime=min_ctime, 
                max_ctime=max_ctime, 
                show_pb=False, 
                session=self.g3session
            )
            self.SMURF.index_timecodes(
                min_ctime=min_ctime, 
                max_ctime=max_ctime, 
                session=self.g3session
            )
            still_missing = len(
                self.SMURF.find_missing_files(timecode, session=self.g3session)
            )
            if still_missing>0:
                msg = f"{still_missing} file(s) were not able to be added to the " \
                    "G3tSmurf database."
                self.logger.error(msg)            
                complete[0] = False
                complete[1] += msg+"\n"
        else:
            self.logger.debug("All files on disk are in G3tSmurf database")

        ## check for level 2 files to be assigned to level 2 observations
        missing_obs = self.SMURF.find_missing_files_from_obs(
            timecode, session=self.g3session
        )
        if len(missing_obs) > 0:
            msg = f"{len(missing_obs)} files not assigned lvl2 obs"
            no_tags = 0
            for fpath in missing_obs:
                if fpath[-6:] != "000.g3":
                    msg += f"\n{fpath} was not added to a larger observation." \
                        " Will be fixed later if possible."
                else:
                    status = SmurfStatus.from_file(fpath)
                    if len(status.tags)==0:
                        no_tags += 1
                    else:
                        msg += f"\Trying to add {fpath} to database"
                        self.SMURF.add_file( 
                            fpath, self.g3session, overwrite=True
                        )
            if no_tags > 0:
                msg += f"\n{no_tags} of the files have no tags, so these should "\
                    "not be observations."
            self.logger.warning(msg)
        
        ## if the stray book has already been bound then we cannot add
        ## more detector books without causing problems
        add_new_detector_books = True
        stray_book = self.session.query(Books).filter( 
            Books.bid == f"stray_{timecode}_{self.platform}",
            Books.status >= BOUND, 
        ).one_or_none()
        if stray_book is not None:
            add_new_detector_books = False

        ## check for incomplete observations
        ## add time to max_ctime to account for observations on the edge
        incomplete = self.imprint._find_incomplete( 
            min_ctime, max_ctime+24*2*3600
        )
        if incomplete.count() > 0:
            ic_list = incomplete.all()
            """Check if these are actually incomplete, imprinter incomplete 
            checker includes making sure the stop isn't beyond max ctime. 
            """
            obs_list = []
            for obs in ic_list:
                if obs.stop is None or obs.timestamp <= max_ctime:
                    obs_list.append(obs)
                    
            ## complete these no matter what for file tracking / deletion 
            self.logger.warning(
                f"Found {len(obs_list)} incomplete observations. Fixing"
            )
            for obs in obs_list:
                self.logger.debug(f"Updating {obs}")
                self.SMURF.update_observation_files(
                    obs, 
                    self.g3session, 
                    force=True,
                )

        ## make sure all obs / operation books from this period are registered
        ## looks like short but overlapping observations are sometimes missed, 
        ## use `try_single_obs` flag to say if we want to try and clean those up
        missing = self.imprint.find_missing_lvl2_obs_from_books(
            min_ctime,max_ctime
        ) 
        if add_new_detector_books and len(missing) > 0:
            self.logger.info(
                f"{len(missing)} lvl2 observations are not registered in books."
                " Trying to register them"
            )
            ## add time to max_ctime to account for observations on the edge
            self.imprint.update_bookdb_from_g3tsmurf(
                min_ctime=min_ctime, max_ctime=max_ctime+24*2*3600,
            )
            still_missing = self.imprint.find_missing_lvl2_obs_from_books(
                min_ctime,max_ctime
            ) 
            if len(still_missing) > 0 and try_single_obs:
                self.logger.warning("Trying single stream registration")
                self.imprint.update_bookdb_from_g3tsmurf(
                    min_ctime=min_ctime, max_ctime=max_ctime+24*2*3600,
                    force_single_stream=True,
                )   
            still_missing = self.imprint.find_missing_lvl2_obs_from_books(
                min_ctime,max_ctime
            ) 
            if len(still_missing) > 0:
                msg = f"Level 2 observations {still_missing} could not be " \
                    "registered in books"
                self.logger.error(msg)
                complete[0] = False
                complete[1] += msg+"\n"
        elif not add_new_detector_books and len(missing)>0:
            msg = f"Have level 2 observations missing but cannot add new " \
                f"detector books because {timecode} was already finalized " \
                " and stray exists. These files should be in stray"
            self.logger.warning(msg)
        
        ## at this point, if an obs or oper book is going to be registered it is
        if try_binding_books:
            books = self.session.query(Books).filter(
                Books.status == UNBOUND,
                Books.start >= dt.datetime.utcfromtimestamp(min_ctime),
                Books.start <= dt.datetime.utcfromtimestamp(max_ctime),
            ).all()
            self.logger.info(f"{len(books)} new books to bind")
            for book in books:
                try:
                    self.imprint.bind_book(book)
                except:
                    self.logger.warning(f"Failed to bind {book.bid}")
        
            failed = self.session.query(Books).filter(
                Books.status == FAILED,
                Books.start >= dt.datetime.utcfromtimestamp(min_ctime),
                Books.start <= dt.datetime.utcfromtimestamp(max_ctime),
            ).all()
            if len(failed) > 0:
                self.logger.info(
                    f"{len(failed)} books failed to bind. trying to autofix"
                )
                autofix_failed_books(
                    self.imprint,
                    min_ctime=min_ctime,
                    max_ctime=max_ctime,
                )
        
        is_final, reason = utils.get_timecode_final(self.imprint, timecode)
        if not is_final:
            self.logger.info(
                f"Timecode {timecode} not counted as final: reason {reason}"
            )
            meta_entries = self.g3session.query(TimeCodes).filter(
                TimeCodes.timecode == timecode,
                TimeCodes.suprsync_type == SupRsyncType.META.value,
            ).count()
            file_entries = self.g3session.query(TimeCodes).filter(
                TimeCodes.timecode == timecode,
                TimeCodes.suprsync_type == SupRsyncType.FILES.value,
            ).count()
            if (
                meta_entries == len(self.imprint.all_slots) and 
                file_entries == len(self.imprint.all_slots)
            ):
                self.logger.info(
                    f"{timecode} was part of the mixed up timecode agent entries"
                )
            elif timecode < tcode_limit:
                self.logger.info(
                    f"At least one server was likely off during timecode {timecode}"
                )
            self.logger.info(
                f"Setting timecode {timecode} to final in SMuRF database"
            )
            utils.set_timecode_final(self.imprint, timecode)

        self.imprint.register_timecode_books(
            min_ctime=min_ctime, 
            max_ctime=max_ctime, 
        )

        if try_binding_books:
            books = self.session.query(Books).filter(
                Books.status == UNBOUND,
                Books.start >= dt.datetime.utcfromtimestamp(min_ctime),
                Books.start <= dt.datetime.utcfromtimestamp(max_ctime),
            ).all()
            self.logger.info(f"{len(books)} new to bind")
            for book in books:
                try:
                    self.imprint.bind_book(book)
                except:
                    self.logger.warning(f"Failed to bind {book.bid}")
        
        # check the smurf book is registered
        book = self.session.query(Books).filter( 
            Books.bid == f"smurf_{timecode}_{self.platform}"
        ).one_or_none()
        if book is None:
            complete[0] = False
            complete[1] += f"SMuRF book smurf_{timecode}_{self.platform} missing\n"
        if include_hk:
            complete = self.check_hk_registered(timecode, complete)
        
        # check if there's a stray book
        stray = self.session.query(Books).filter( 
            Books.bid == f"stray_{timecode}_{self.platform}"
        ).one_or_none()
        if stray is None and try_binding_books:
            # all files should be in obs/oper books
            flist = self.imprint.get_files_for_stray_book(
                min_ctime=min_ctime, 
                max_ctime=max_ctime, 
            )
            if len(flist) > 0:
                complete[0] = False
                complete[1] += f"Stray book stray_{timecode}_{self.platform} missing\n"
        elif stray is None and not try_binding_books:
            my_list = self.imprint.get_files_for_stray_book(
                min_ctime=min_ctime, 
                max_ctime=max_ctime, 
            )
            if len(my_list) > 0:
                self.logger.warning(
                    f"We expect {len(my_list)} books in a stray book but need "
                    "to bind books to verify"
                )
                complete[0] = False
                complete[1] += f"Stray book stray_{timecode}_{self.platform} missing\n"
        else:
            flist = self.imprint.get_files_for_book(stray)
            my_list = self.imprint.get_files_for_stray_book(
                min_ctime=min_ctime, 
                max_ctime=max_ctime, 
            )
            assert np.all(
                sorted(flist) == sorted(my_list)
            ), "logic error somewhere"
        ## check that all books are bound
        books = self.session.query(Books).filter(
            or_(Books.status == UNBOUND, Books.status == FAILED),
            Books.start >= dt.datetime.utcfromtimestamp(min_ctime),
            Books.start <= dt.datetime.utcfromtimestamp(max_ctime),
        ).count() 
        if books != 0:
            complete[0] = False
            complete[1] += f"Have {books} unbound or failed books in timecode \n"
        return complete

    def books_in_timecode(
        self, timecode, include_wont_fix=False, include_hk=True
    ):
        min_ctime = timecode*1e5
        max_ctime = (timecode+1)*1e5

        q = self.session.query(Books).filter(
            Books.start >= dt.datetime.utcfromtimestamp(min_ctime),
            Books.start < dt.datetime.utcfromtimestamp(max_ctime),
        )
        if not include_wont_fix:
            q = q.filter(Books.status != WONT_BIND)
        if not include_hk:
            q = q.filter(Books.type != 'hk')
        return q.all()

    def file_list_from_database(
        self, timecode, deletable, verify_with_librarian, include_hk=True,
    ):
        file_list = []
        min_ctime = timecode*1e5
        max_ctime = (timecode+1)*1e5

        q = self.session.query(Books).filter(
            Books.start >= dt.datetime.utcfromtimestamp(min_ctime),
            Books.start < dt.datetime.utcfromtimestamp(max_ctime),
        )
        not_ready = q.filter( not_(or_( 
            Books.status == WONT_BIND, Books.status >= UPLOADED)
        )).count()
        if not_ready > 0:
            self.logger.error(
                f"There are {not_ready} non-uploaded books in this timecode"
            )
            deletable[0] = False
            deletable[1] += f"There are {not_ready} non-uploaded books in " \
                "this timecode\n"
        if not include_hk:
            q = q.filter(Books.type != 'hk')
        book_list = q.filter(Books.status >= UPLOADED).all()
        self.logger.debug(
            f"Found {len(book_list)} books in time code {timecode}"
        )

        for book in book_list:
            if book.lvl2_deleted:
                continue
            if verify_with_librarian:
                in_lib = self.imprint.check_book_in_librarian(
                    book, n_copies=1, raise_on_error=False
                )
                if not in_lib:
                    deletable[0] = False
                    deletable[1] += f"{book.bid} has not been uploaded to librarain\n"

            flist = self.imprint.get_files_for_book(book)
            if isinstance(flist, OrderedDict):
                x = []
                for k in flist:
                    x.extend(flist[k])
                flist=x
            ## annoying edge case catcher because we use datetimes in
            ## book.start and book.stop
            for f in flist:
                if f"/{timecode}/" not in f:
                    self.logger.info(
                        f"Not adding {f} to file list because it is not in "
                        f"the {timecode} folder"
                    )
                    continue
                file_list.append(f)
        # add suprsync files 
        file_list.extend( self.get_suprsync_files(timecode) )
        return file_list, deletable

    def verify_timecode_deletable(
        self, timecode, verify_with_librarian=True, include_hk=True,
    ):
        """
        Checkes that all books in that timecode are uploaded to the librarian 
        and that there is a copy offsite (if verify_with_librarian=True)

        Steps for checking:

        1. Walk the file system and build up a list of all files there
        2. Go book by book within timecode and build up the list of level 2 
        files that went into it using the databases. Add any files in suprsync 
        folders into this list since they aren't book bound but we'd like them 
        to be deleted 
        3. Compare the two lists and make sure they're the same.
        """
        deletable = [True, ""]

        files_on_disk = self.all_files_in_timecode(
            timecode, include_hk=include_hk
        )
        if len(files_on_disk) == 0:
            return deletable
        # these are files that are in the smurf directory but we don't save in the 
        # smurf books. mostly watching out for .dat files
        ignore = shutil.ignore_patterns(*SMURF_EXCLUDE_PATTERNS)
        ignored_files = ignore("", files_on_disk)
        self.logger.debug(
            f"Timecode {timecode} has {len(ignored_files)} ignored files"
        )
        files_in_database, deletable = self.file_list_from_database(
            timecode, deletable, verify_with_librarian, include_hk=include_hk
        )

        missed_files = []
        extra_files = []
        for f in files_on_disk:
            if f not in files_in_database and f not in ignored_files:
                missed_files.append(f)
        for f in files_in_database:
            if f not in files_on_disk:
                extra_files.append(f)
        if len(missed_files) == 0 and len(extra_files) == 0:
            self.logger.info(f"Timecode {timecode} has complete coverage")
        if len(missed_files)>0:
            msg = f"Files on disk but not marked as uploaded in database"
            msg += f" {len(missed_files)}:\n"
            for f in missed_files:
                msg += f"\t{f}\n"
            self.logger.warning(msg)
            deletable[0] = False
            deletable[1] += msg
        if len(extra_files)>0:
            msg = f"Files in database but not on disk: {extra_files}"
            for f in missed_files:
                msg += f"\t{f}\n"
            self.logger.error(msg)
            deletable[0] = False
            deletable[1] += msg
        return deletable

    def delete_timecode_level2(
        self, timecode, dry_run=True, include_hk=True, 
        verify_with_librarian=True,
    ):
        book_list = self.books_in_timecode(timecode, include_hk=include_hk)
        books_not_deleted = []

        for book in book_list:
            stat = self.imprint.delete_level2_files(
                book, verify_with_librarian=verify_with_librarian,
                n_copies_in_lib=2, dry_run=dry_run, n_tries=2
            )
            if stat > 0:
                books_not_deleted.append(book)    
        
        if len(books_not_deleted) > 0:
            msg = "Could not delete level 2 for books:\n"
            for book in books_not_deleted:
                msg += f'\t{book.bid}\n'   
            self.logger.error(msg)
            return False, ""
        return True, ""

    
    def delete_timecode_staged(
        self, timecode, include_hk=True, verify_with_librarian=False,
        check_level2=False,
    ):
        book_list = self.books_in_timecode(timecode, include_hk=include_hk)
        books_not_deleted = []
        for book in book_list:
            stat = self.imprint.delete_book_staged(
                book, check_level2=check_level2, 
                verify_with_librarian=verify_with_librarian
            )
            if stat > 0:
                books_not_deleted.append(book)        
        # cleanup
        for tube in self.imprint.tubes:
            for btype in ['obs', 'oper']:
                path = os.path.join(
                    self.imprint.output_root, tube, btype, str(timecode)
                )
                if os.path.exists(path) and len(os.listdir(path))==0:
                    os.rmdir(path)
        
        if len(books_not_deleted) > 0:
            msg = "Could not delete stages for books:\n"
            for book in books_not_deleted:
                msg += f'\t{book.bid}\n'   
            self.logger.error(msg)
            return False, msg
        return True, ""
    
    def check_and_delete_timecode(
        self, timecode, include_hk=True, verify_with_librarian=True
    ):
        check = self.make_timecode_complete(timecode, include_hk=include_hk)
        if not check[0]:
            self.logger.error(f"Timecode {timecode} not complete")
            self.logger.error(check[1])
            return check
        check = self.verify_timecode_deletable(
            timecode, include_hk=include_hk, 
            verify_with_librarian=False,
        )
        if not check[0]:
            self.logger.error(f"Timecode {timecode} not ready to delete")
            self.logger.error(check[1])
            return check
        
        check = self.delete_timecode_level2(
            timecode, dry_run=False, include_hk=include_hk,
            verify_with_librarian=verify_with_librarian,
        )

        if not self.imprint.build_det:
            return check
        stc = os.path.join(self.SMURF.meta_path, str(timecode))
        ttc = os.path.join(self.SMURF.archive_path, str(timecode))

        if os.path.exists(stc): 
            if len(os.listdir(stc)) == 0 or just_suprsync(stc):
                shutil.rmtree(stc)
        if os.path.exists(ttc):
            if len(os.listdir(ttc)) == 0 or just_suprsync(ttc):
                shutil.rmtree(ttc)
        return check



import sqlite3
import os
import numpy as np
import warnings

from .resultset import ResultSet
from . import common
from .. import util


TABLE_DEFS = {
    'obs': [
        "`obs_id` varchar(256) primary key",
        "`timestamp` float",
    ],
    'tags': [
        "`obs_id` varchar(256)",
        "`tag` varchar(256)",
        "CONSTRAINT one_tag UNIQUE (`obs_id`, `tag`)",
    ],
    '_indices': {
        'idx_obs': 'obs(obs_id)',
        'idx_tags': 'tags(obs_id)',
    },
}


class ObsDb(object):
    """Observation database.

    The ObsDb helps to associate observations, indexed by an obs_id,
    with properties of the observation that might be useful for
    selecting data or for identifying metadata.

    The main ObsDb table is called 'obs', and contains the columns
    obs_id (string), plus any others deemed important for this context
    (you will probably find timestamp (float representing a unix
    timestamp)).  Additional columns may be added to this table as
    needed.

    The second ObsDb table is called 'tags', and facilitates grouping
    observations together using string labels.

    """

    TABLE_TEMPLATE = [
        "`obs_id` varchar(256)",
    ]

    def __init__(self, map_file=None, init_db=True):
        """Instantiate an ObsDb.

        Args:
          map_file (str or sqlite3.Connection): If this is a string,
            it will be treated as the filename for the sqlite3
            database, and opened as an sqlite3.Connection.  If this is
            an sqlite3.Connection, it is cached and used.  If this
            argument is None (the default), then the
            sqlite3.Connection is opened on ':memory:'.
          init_db (bool): If True, then any ObsDb tables that do not
            already exist in the database will be created.

        Notes:
          If map_file is provided, the database will be connected to
          the indicated sqlite file on disk, and any changes made to
          this object be written back to the file.

        """
        if isinstance(map_file, sqlite3.Connection):
            self.conn = map_file
        else:
            if map_file is None:
                map_file = ':memory:'
            self.conn = sqlite3.connect(map_file)

        self.conn.row_factory = sqlite3.Row  # access columns by name
        if init_db:
            c = self.conn.cursor()
            c.execute("SELECT type, name FROM sqlite_master "
                      "WHERE type in ('table', 'index') and name not like 'sqlite_%';")
            tables = [r[1] for r in c]
            changes = False
            for k, v in TABLE_DEFS.items():
                if k[0] != '_' and k not in tables:
                    q = ('create table if not exists `%s` (' % k +
                         ','.join(v) + ')')
                    c.execute(q)
                    changes = True
            for index, cols in TABLE_DEFS['_indices'].items():
                if index not in tables:
                    c.execute(f'CREATE INDEX IF NOT EXISTS {index} on {cols}')
                    changes = True
            if changes:
                self.conn.commit()

    def __len__(self):
        return self.conn.execute('select count(obs_id) from obs').fetchone()[0]

    def add_obs_columns(self, column_defs, ignore_duplicates=True, commit=True):
        """Add columns to the obs table.

        Args:
          column_defs (list of pairs of str): Column descriptions, see
            notes.
          ignore_duplicates (bool): If true, requests for new columns
            will be ignored if the column name is already present in
            the table.

        Returns:
          self.

        Notes:
          The input format for column_defs is somewhat flexible.
          First of all, if a string is passed in, it will converted to
          a list by splitting on ",".  Second, if the items in the
          list are strings (rather than tuples), the string will be
          broken into 2 components by splitting on whitespace.
          Finally, each pair of items is interpreted as a (name, data
          type) pair.  The name can be a simple string, or a string
          inside backticks; so 'timestamp' and '`timestamp`' are
          equivalent.  The data type can be any valid sqlite type
          expression (e.g. 'float', 'varchar(256)', etc) or it can be
          one of the three basic python type objects: str, float, int.
          Here are some examples of valid column_defs arguments::

            [('timestamp', float), ('drift', str)]
            ['`timestamp` float', '`drift` varchar(32)']
            'timestamp float, drift str'

        """
        current_cols = self.conn.execute('pragma table_info(obs)').fetchall()
        current_cols = [r[1] for r in current_cols]
        if isinstance(column_defs, str):
            column_defs = column_defs.split(',')
        for column_def in column_defs:
            if isinstance(column_def, str):
                column_def = column_def.split()
            name, typestr = column_def
            if typestr is float:
                typestr = 'float'
            elif typestr is int:
                typestr = 'int'
            elif typestr is str:
                typestr = 'text'
            check_name = name
            if name.startswith('`'):
                check_name = name[1:-1]
            else:
                name = '`' + name + '`'
            if check_name in current_cols:
                if ignore_duplicates:
                    continue
                raise ValueError("Column %s already exists in table obs" % check_name)
            self.conn.execute('ALTER TABLE obs ADD COLUMN %s %s' % (name, typestr))
            current_cols.append(check_name)
        if commit:
            self.conn.commit()
        return self

    def update_obs(self, obs_id, data={}, tags=[], commit=True):
        """Update an entry in the obs table.

        Arguments:
          obs_id (str): The id of the obs to update.
          data (dict): map from column_name to value.
          tags (list of str): tags to apply to this observation (if a
            tag name is prefxed with '!', then the tag name will be
            un-applied, i.e. cleared from this observation.

        Returns:
            self.

        """
        c = self.conn.cursor()
        c.execute('INSERT OR IGNORE INTO obs (obs_id) VALUES (?)',
                  (obs_id,))
        if len(data.keys()):
            settors = [f'{k}=?' for k in data.keys()]
            c.execute('update obs set ' + ','.join(settors) + ' '
                      'where obs_id=?',
                      tuple(data.values()) + (obs_id, ))
        for t in tags:
            if t[0] == '!':
                # Kill this tag.
                c.execute('DELETE FROM tags WHERE obs_id=? AND tag=?',
                          (obs_id, t[1:]))
            else:
                c.execute('INSERT OR REPLACE INTO tags (obs_id, tag) '
                          'VALUES (?,?)', (obs_id, t))
        if commit:
            self.conn.commit()
        return self

    def copy(self, map_file=None, overwrite=False):
        """
        Duplicate the current database into a new database object, and
        return it.  If map_file is specified, the new database will be
        connected to that sqlite file on disk.  Note that a quick way
        of writing a Db to disk to call copy(map_file=...) and then
        simply discard the returned object.
        """
        if map_file is not None and os.path.exists(map_file):
            if overwrite:
                os.remove(map_file)
            else:
                raise RuntimeError("Output file %s exists (overwrite=True "
                                   "to overwrite)." % map_file)
        new_db = ObsDb(map_file=map_file, init_db=False)
        script = ' '.join(self.conn.iterdump())
        new_db.conn.executescript(script)
        return new_db

    def to_file(self, filename, overwrite=True, fmt=None):
        """Write the present database to the indicated filename.

        Args:
          filename (str): the path to the output file.
          overwrite (bool): whether an existing file should be
            overwritten.
          fmt (str): 'sqlite', 'dump', or 'gz'.  Defaults to 'sqlite'
            unless the filename ends with '.gz', in which it is 'gz'.

        """
        return common.sqlite_to_file(self.conn, filename, overwrite=overwrite, fmt=fmt)

    @classmethod
    def from_file(cls, filename, fmt=None, force_new_db=True):
        """This method calls
            :func:`sotodlib.core.metadata.common.sqlite_from_file`
        """
        conn = common.sqlite_from_file(filename, fmt=fmt, force_new_db=force_new_db)
        return cls(conn, init_db=False)

    def get(self, obs_id=None, tags=None, add_prefix=''):
        """Returns the entry for obs_id, as an ordered dict.

        If obs_id is None, returns all entries, as a ResultSet.
        However, this usage is deprecated in favor of self.query().

        Args:
          obs_id (str): The observation id to get info for.
          tags (bool): Whether or not to load and return the tags.
          add_prefix (str): A string that will be prepended to each
            field name.  This is for the lazy metadata system, because
            obsdb selectors are prefixed with 'obs:'.

        Returns:
          An ordered dict with the obs table entries for this obs_id,
          or None if the obs_id is not found.  If tags have been
          requested, they will be stored in 'tags' as a list of strings.

        """
        if obs_id is None:
            return self.query('1', add_prefix=add_prefix)
        results = self.query(f"obs_id='{obs_id}'", add_prefix=add_prefix)
        if len(results) == 0:
            return None
        if len(results) > 1:
            raise ValueError('Too many rows...')  # or integrity error...
        output = results[0]
        if tags:
            # "distinct" should not be needed given uniqueness constraint.
            c = self.conn.execute('select distinct tag from tags where obs_id=?', (obs_id,))
            output['tags'] = [r[0] for r in c]
        return output

    def query(self, query_text='1', tags=None, sort=['obs_id'], add_prefix=''):
        """Queries the ObsDb using user-provided text.  Returns a ResultSet.

        Args:
          query_text (str): The sqlite query string.  All fields
            should refer to the obs table, or to tags explicitly
            listed in the tags argument.
          tags (list of str): Tags to include in the output; if they
            are listed here then they can also be used in the query
            string.  Filtering on tag value can be done here by
            appending '=0' or '=1' to a tag name.

        Returns:
          A ResultSet with one row for each Observation matching the
          criteria.

        Notes:
          Tags are added to the output on request.  For example,
          passing tags=['planet','stare'] will cause the output to
          include columns 'planet' and 'stare' in addition to all the
          columns defined in the obs table.  The value of 'planet' and
          'stare' in each row will be 0 or 1 depending on whether that
          tag is set for that observation.  We can include expressions
          involving planet and stare in the query, for example::

            obsdb.query('planet=1 or stare=1', tags=['planet', 'stare'])

          For simple filtering on tags, pass '=1' or '=0', like this::

            obsdb.query(tags=['planet=1','hwp=1'])

          When filtering is activated in this way, the returned
          results must satisfy all the criteria (i.e. the individual
          constraints are AND-ed).

        """
        sort_text = ''
        if sort is not None and len(sort):
            sort_text = ' ORDER BY ' + ','.join(sort)
        if '"' in query_text:
            warnings.warn('obsdb.query text contains double quotes (") -- '
                          'replacing with single quotes (\').')
            query_text = query_text.replace('"', "'")

        joins = ''
        extra_fields = []
        if tags is not None and len(tags):
            for tagi, t in enumerate(tags):
                if '=' in t:
                    t, val = t.split('=')
                else:
                    val = None
                if val is None:
                    join_type = 'left join'
                    extra_fields.append(f"ifnull(tt{tagi}.obs_id,'') != '' as {t}")
                elif val == '0':
                    join_type = 'left join'
                    extra_fields.append(f"ifnull(tt{tagi}.obs_id,'') != '' as {t}")
                    query_text += f' and {t}==0'
                else:
                    join_type = 'join'
                    extra_fields.append(f'1 as {t}')
                joins += (f" {join_type} (select distinct obs_id from tags where tag='{t}') as tt{tagi} on "
                          f"obs.obs_id = tt{tagi}.obs_id")
        extra_fields = ''.join([','+f for f in extra_fields])
        q = 'select obs.* %s from obs %s where %s %s' % (extra_fields, joins, query_text, sort_text)
        c = self.conn.execute(q)
        results = ResultSet.from_cursor(c)
        if add_prefix is not None:
            results.keys = [add_prefix + k for k in results.keys]
        return results

    def info(self):
        """Return a dict summarizing the structure and contents of the obsdb;
        this is used by the CLI.

        """
        def _short_list(items, max_len=40):
            i, acc, keepers = 0, 0, []
            while (len(keepers) < 1 or acc < max_len) and i < len(items):
                keepers.append(str(items[i]))
                i += 1
                acc += len(keepers[-1]) + 2
            return  ('[' + ', '.join(map(str, keepers))
                     + (' ...' * (i < len(items))) + ']')

        # Summarize the fields ...
        rs = self.query()
        fields = {}
        for k in rs.keys:
            items = list(set(rs[k]))
            fields[k] = (len(items), _short_list(items))

        # Count occurances of each tag ...
        c = self.conn.execute('select tag, count(obs_id) from tags group by tag order by tag')
        tags = {r[0]: r[1] for r in c}

        return {
            'count': len(rs),
            'fields': fields,
            'tags': tags,
        }


def diff_obsdbs(obsdb_left, obsdb_right, return_detail=False):
    """Examine all records in two obsdbs and construct a list of changes
    that could made to obsdb_left in order to make it match
    obsdb_right.

    Returns a dict with following entries:

    - ``different`` (bool): whether the two databases carry different
      information.
    - ``patchable`` (bool): whether the function was able to construct
      patching instructions.
    - ``unpatchable_reason`` (str): if not patchable, a string
      explaining why.
    - ``detail`` (various): if not patchable, and return_detail, then
      this will contain detail about the offending data (e.g. obs rows
      in the two dbs that contain discrepant data).
    - ``patch_data`` (dict): if patchable, the data needed to patch
      obsdb_left.  The fields are:

      - ``remove_obs`` (list of obs_id): entries to remove from obs
        table.
      - ``remove_tags`` (list of tuple): entries to remove from tags
        table.
      - ``new_obs`` (list of dict): rows of new data for obs table --
        each dict can be passed directly to obsdb.update_obs.
      - ``new_tags`` (list of tuple): rows of data for tags table
        (obs_id, tag).


    Notes:

    In the present implementation, only changes involving adding rows
    to obsdb_left (either whole obs rows or tag rows) will yield a
    patchable result.  Cases where some data has changed, or obs or
    tags have been deleted, will simply return as unpatchable.  This
    is probably pretty easy to extend, should the need arise.

    """
    if isinstance(obsdb_left, str):
        obsdb_left = ObsDb.from_file(obsdb_left, force_new_db=False)
    if isinstance(obsdb_right, str):
        obsdb_right = ObsDb.from_file(obsdb_right, force_new_db=False)

    def failure_declaration(reason, detail=None):
        if not return_detail:
            detail = None
        return {'different': True,
                'patchable': False,
                'unpatchable_reason': reason,
                'detail': detail}

    full = [db.query() for db in [obsdb_left, obsdb_right]]
    if full[0].keys != full[1].keys:
        return failure_declaration(
            'obsdb_left and obsdb_right have different column names.',
            detail=[full[0].keys, full[1].keys])

    # Convert to arrays.
    obs_ids = [set(f['obs_id']) for f in full]

    # Insist right is superset of left.
    left_not_right = sorted(list(obs_ids[0].difference(obs_ids[1])))
    if len(left_not_right):
        return failure_declaration(
            f'obsdb_left contains {len(left_not_right)} '
            'obs not found in obsdb_right.',
            detail=left_not_right)

    # Any obs in common?
    unmatched_right = np.ones(len(full[1]), bool)
    common = sorted(list(obs_ids[0].intersection(obs_ids[1])))
    if len(common):
        common, i0, i1 = util.get_coindices(*(f['obs_id'] for f in full))
        diffs = []
        for i, (_i0, _i1) in enumerate(zip(i0, i1)):
            if full[0][_i0] != full[1][_i1]:
                diffs.append((full[0][_i0], full[1][_i1]))
        if len(diffs):
            return failure_declaration(
                f'obsdb_left and obsdb_right have {len(diffs)} obs '
                'in common, with different data.',
                detail=diffs)
        unmatched_right[i1] = False

    # Ok finally
    pd = {
        'remove_obs': [],
        'remove_tags': [],
        'new_obs': [],
        'new_tags': [],
    }
    for idx in unmatched_right.nonzero()[0]:
        pd['new_obs'].append(full[1][idx])

    # Tag check.
    tags_tuples = [
        list(map(tuple, db.conn.execute(
            'select distinct obs_id, tag from tags  '
            'order by obs_id, tag').fetchall()))
        for db in [obsdb_left, obsdb_right]]

    # Collapse tags to single strings and eliminate duplicates.
    DELIM = ':::/:::'
    common, i0, i1 = util.get_coindices(*[[t[0] + DELIM + t[1] for t in tt]
                                          for tt in tags_tuples])
    if len(i0) != len(tags_tuples[0]):
        return failure_declaration(
            f'obsdb_left contains {len(tags_tuples[0]) - len(i0)} '
            'tags not found in obsdb_right',
            detail=list(set(tags_tuples[0]).difference(tags_tuples[1])))
    unmatched_right = np.ones(len(tags_tuples[1]), bool)
    unmatched_right[i1] = False
    for idx in unmatched_right.nonzero()[0]:
        pd['new_tags'].append(tags_tuples[1][idx])

    different = any([(len(v) != 0) for v in pd.values()])

    return {
        'different': different,
        'patchable': True,
        'patch_data': pd,
    }


def patch_obsdb(patch_data, target_db):
    """Update an ObsDb with a batch of changes.

    Args:
      target_db (ObsDb): the database where changes should be made.
      patch_data (dict): patch information, as returned by
        diff_obsdbs.

    """
    assert len(patch_data['remove_obs']) == 0
    assert len(patch_data['remove_tags']) == 0

    for obs_entry in patch_data['new_obs']:
        target_db.update_obs(obs_entry['obs_id'], obs_entry,
                             commit=False)

    # Group new tags by obs.
    tags_obsed = {}
    for k, v in patch_data['new_tags']:
        if k not in tags_obsed:
            tags_obsed[k] = [v]
        else:
            tags_obsed[k].append(v)
    for obs, tags in tags_obsed.items():
        target_db.update_obs(obs, {}, tags=tags, commit=False)

    target_db.conn.commit()
